#!/usr/bin/env python3

import re
from pathlib import Path
from typing import Annotated, Dict, Optional

from fastapi import (
    APIRouter,
    Body,
    Depends,
    File,
    HTTPException,
    Query,
    UploadFile,
)
from fastapi.responses import FileResponse
from loguru import logger

from baobaxia.api.v2.endpoints.auth import get_current_mocambola
from baobaxia.constants import (
    ROLE_ACERVO_EDITOR,
    ROLE_ACERVO_PUBLISHER,
    SABER_FILE_EXT,
)
from baobaxia.models.artigo import Artigo, ArtigoStatus
from baobaxia.permissions.base import PermissionChecker
from baobaxia.permissions.model_permissions import Artigos
from baobaxia.root import Baobaxia, bbx
from baobaxia.saberes import Mocambola
from baobaxia.util import calculate_md5, discover_mimetype, norm
from baobaxia.utils.strings import parse_tags

router = APIRouter(tags=["Artigos"])
CurrentMocambola = Annotated[Mocambola, Depends(get_current_mocambola)]


@router.post(
    "/{balaio_slug_smid}/{mucua_slug_smid}/blog/artigo",
    #    dependencies=[Depends(PermissionChecker([Artigos.permissions.CREATE]))],
    summary="Cria um artigo",
)
async def post_artigo(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    mocambola: CurrentMocambola,
    titulo: str = Body(...),
    description: str = Body(...),
    autor: str = Body(default=None),
    mocambo: str = Body(default=None),
    is_public: bool = Body(default=True),
    tags: str = Body(default=None),
) -> Artigo:
    """Cria um saber artigo"""
    logger.info("Criando um artigo")
    balaio_smid, mucua_smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid, mucua_slug_smid=mucua_slug_smid
    )

    tags_list = parse_tags(tags)

    artigo = Artigo(
        balaio_smid=balaio_smid,
        mucua_smid=mucua_smid,
        name=titulo,
        titulo=titulo,
        description=description,
        autor=autor,
        mocambo=mocambo,
        is_public=is_public,
        tags=tags_list,
    )

    logger.info("Modelo de Artigo criado: {artigo}")
    return bbx.put_artigo(balaio_smid, mucua_smid, artigo, mocambola)


@router.put(
    "/{balaio_slug_smid}/{mucua_slug_smid}/blog/artigo/{slug_smid:str}",
    # dependencies=[Depends(PermissionChecker([Artigos.permissions.UPDATE]))],
    summary="Atualiza um artigo",
)
async def put_artigo(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    saber: Artigo,
    mocambola: CurrentMocambola,
) -> Artigo:
    """Atualiza um artigo"""
    logger.debug("Call put_artigo function...")
    balaio_smid, mucua_smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid, mucua_slug_smid=mucua_slug_smid
    )
    return bbx.put_artigo(balaio_smid, mucua_smid, saber, mocambola)


@router.post(
    "/{balaio_slug_smid}/{mucua_slug_smid}/blog/upload/{slug_smid:str}"
)
async def upload_artigo(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    slug_smid: str,
    mocambola: CurrentMocambola,
    arquivo: UploadFile = File(...),
) -> Dict:
    """Enviar artigo já existente"""
    balaio_smid, mucua_smid, smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid,
        mucua_slug_smid=mucua_slug_smid,
        slug_smid=slug_smid,
    )
    saber = bbx.get_artigo(balaio_smid, mucua_smid, smid)

    tmp_file = f"/tmp/{arquivo.filename}"
    with open(tmp_file, "wb") as buffer:
        buffer.write(arquivo.file.read())
        file_size = Path(tmp_file).stat().st_size
        extension = Path(tmp_file).suffix
    hash_sum = f"MD5E-s{file_size}--{calculate_md5(tmp_file)}{extension}"

    arquivo.file.seek(0)

    logger.debug(f"FILE HASHSUM: {hash_sum}")
    logger.debug(f"FILE SIZE: {file_size}")

    content = {
        "filename": arquivo.filename,
        "hash_sum": hash_sum,
        "mimetype": discover_mimetype(arquivo.filename),
    }
    saber.content.append(content)
    with (
        bbx.get_balaio_mucua_path(balaio_smid, mucua_smid)
        / saber.path
        / arquivo.filename
    ).open("wb") as arquivo_saber:
        arquivo_saber.write(arquivo.file.read())
        arquivo_saber.close()
        bbx.put_artigo(balaio_smid, mucua_smid, saber, mocambola)
        saber._is_local = True
        saber_path = (
            bbx.get_balaio_mucua_path(balaio_smid, mucua_smid)
            / saber.path
            / SABER_FILE_EXT
        )
        saber_path.with_suffix(".local").open("w").write(
            saber.model_dump_json(indent=4)
        )
    return {"detail": "success"}


@router.get("/blog/find")
async def find_artigos(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: Optional[str] = Query(None),
    mucua_slug_smid: Optional[str] = Query(None),
    mocambola: Optional[CurrentMocambola] = None,
    keywords: Optional[str] = Query(None),
    hashtags: Optional[str] = Query(None),
    status: Optional[str] = Query(None),
    creator: Optional[str] = Query(None),
    ordem_campo: Optional[str] = Query(None),
    ordem_decrescente: bool = Query(False),
    pag_tamanho: int = 12,
    pag_atual: int = 1,
) -> Dict:
    """Busca artigos de acordo com os parâmetros fornecidos."""
    if balaio_slug_smid is not None and balaio_slug_smid:
        balaio_smid = bbx.extract_smid(balaio_slug_smid)
    else:
        balaio_smid = None
    if mucua_slug_smid is not None and mucua_slug_smid:
        mucua_smid = bbx.extract_smid(mucua_slug_smid)
    else:
        mucua_smid = None
    kw_list = []
    if keywords is not None and len(keywords) > 0:
        kw_list = keywords.split(" ")
    ht_list = []
    if hashtags is not None and len(hashtags) > 0:
        # ht_list = hashtags.split(' ')
        ht_list = re.split("; |, ", hashtags)

    def filter_function(artigo):
        if status is not None and norm(artigo.status) != norm(status):
            return False
        if creator is not None and norm(artigo.creator) != norm(creator):
            return False
        match = True
        for kw in kw_list:
            if (
                norm(kw) not in norm(artigo.titulo)
                and norm(kw) not in norm(artigo.texto)
                and (
                    norm(kw) not in norm(artigo.resumo)
                    if "resumo" in artigo
                    else True
                )
            ):
                match = False
                break
        for ht in ht_list:
            if norm(ht) not in norm(artigo.tags):
                match = False
                break
        return match

    def sorted_function(artigo):
        if ordem_campo is None:
            return 0
        elif hasattr(artigo, ordem_campo):
            return getattr(artigo, ordem_campo)
        else:
            return 0

    return bbx.find_artigos(
        balaio_smid,
        mucua_smid,
        mocambola,
        filter_function=filter_function,
        sorted_function=sorted_function,
        sorted_reverse=ordem_decrescente,
        page_size=pag_tamanho,
        page_index=pag_atual,
    )


@router.get(
    "/{balaio_slug_smid}/{mucua_slug_smid}/blog/artigo/{slug_smid}",
)
async def get_artigo(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    slug_smid: str,
) -> Artigo:
    """Retorna um Artigo."""
    balaio_smid, mucua_smid, smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid,
        mucua_slug_smid=mucua_slug_smid,
        slug_smid=slug_smid,
    )
    return bbx.get_artigo(balaio_smid, mucua_smid, smid)


@router.delete(
    "/{balaio_slug_smid}/{mucua_slug_smid}/blog/artigo/{slug_smid:str}"
)
async def delete_artigo(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    slug_smid: str,
    mocambola: CurrentMocambola,
) -> Dict:
    """Remove um saber artigo"""
    balaio_smid, mucua_smid, smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid,
        mucua_slug_smid=mucua_slug_smid,
        slug_smid=slug_smid,
    )
    bbx.del_artigo(balaio_smid, mucua_smid, smid, mocambola)
    return {"detail": "success"}


@router.get(
    "/{balaio_slug_smid}/{mucua_slug_smid}/blog/download/{slug_smid:str}"
)
@router.get(
    "/{balaio_slug_smid}/{mucua_slug_smid}/blog/download/{slug_smid:str}/{index}"
)
async def download_arquivo(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    slug_smid: str,
    index: Optional[int] = 0,
):
    """Retornar arquivo markdown ou de mídia"""
    balaio_smid, mucua_smid, smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid,
        mucua_slug_smid=mucua_slug_smid,
        slug_smid=slug_smid,
    )
    saber = bbx.get_artigo(balaio_smid, mucua_smid, smid)
    index = index if index > 0 else 0
    file_list = [item["filename"] for item in saber.content]
    if index > len(saber.content) or len(saber.content) == 0:
        raise HTTPException(status_code=404, detail="Acervo não encontrado")
    return FileResponse(
        path=str(
            bbx.get_balaio_mucua_path(balaio_smid, mucua_smid)
            / saber.path
            / file_list[index]
        )
    )
