#!/usr/bin/env python3

"""Grupo de rotas para autenticação."""
from typing import Annotated, Dict, List, Optional

from fastapi import APIRouter, Body, Depends, File, Form, Header, UploadFile
from fastapi.exceptions import HTTPException
from loguru import logger

from baobaxia.api.v2.endpoints.auth import (
    PermissionChecker,
    get_current_mocambola,
)
from baobaxia.constants import ROLE_MUCUA_PAJE
from baobaxia.permissions.model_permissions import *
from baobaxia.root import Baobaxia, bbx
from baobaxia.saberes import Mocambola, Mucua

CurrentMocambola = Annotated[Mocambola, Depends(get_current_mocambola)]
router = APIRouter(tags=["Mucuas"])


@router.get("/mucua/default")
async def get_default_mucua(
    bbx: Annotated[Baobaxia, Depends(bbx)],
) -> Mucua:
    """Retorn a mucua padrão."""
    return bbx.get_default_mucua()


@router.get("/{balaio_slug_smid}/mucua")
async def list_mucuas(
    bbx: Annotated[Baobaxia, Depends(bbx)], balaio_slug_smid: str
) -> List[Mucua]:
    """Listar mucuas."""
    balaio_smid = bbx.extract_smid(balaio_slug_smid=balaio_slug_smid)
    return bbx.list_mucuas(balaio_smid)


@router.get("/mucua")
async def list_all_mucuas(
    bbx: Annotated[Baobaxia, Depends(bbx)],
) -> List[Mucua]:
    """Listar todas as mucuas de todos os balaios."""
    return bbx.list_mucuas()


@router.post("/{balaio_slug_smid}/mucua")
async def post_mucua(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mocambola: CurrentMocambola,
    name: str = Body(...),
    description: str = Body(default=None),
    creator: str = Body(default=None),
    is_public: bool = Body(default=True),
) -> Mucua:
    """Criar nova mucua."""

    balaio_smid = bbx.extract_smid(balaio_slug_smid=balaio_slug_smid)

    mucua = Mucua(
        name=name,
        balaio_smid=balaio_smid,
        description=description,
        creator=creator,
        is_public=is_public,
    )
    logger.info("Criando nova Mucua!")

    return bbx.post_mucua(
        balaio_smid=balaio_smid,
        mucua=mucua,
        mocambola=mocambola,
        is_public=mucua.is_public,
    )


@router.get("/{balaio_slug_smid}/mucua/{mucua_slug_smid}")
async def get_mucua(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
) -> Mucua:
    """Retornar uma mucua."""
    balaio_smid, mucua_smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid, mucua_slug_smid=mucua_slug_smid
    )
    return bbx.get_mucua(balaio_smid, mucua_smid)


@router.put("/{balaio_slug_smid}/mucua/{mucua_slug_smid}")
async def put_mucua(
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    mocambola: CurrentMocambola,
    mucua: Mucua,
) -> Mucua:
    """Atualizar uma mucua."""
    balaio_smid, mucua_smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid, mucua_slug_smid=mucua_slug_smid
    )
    return bbx.put_mucua(balaio_smid, mucua_smid, mocambola, mucua)


@router.put("/{balaio_slug_smid}/mucua/{mucua_slug_smid}/upload_image")
async def upload_mucua_image(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    mocambola: CurrentMocambola,
    arquivo: UploadFile = File(...),
) -> Dict:
    """Enviar a imagem da mucua."""
    balaio_smid, mucua_smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid, mucua_slug_smid=mucua_slug_smid
    )
    mocambola = bbx.get_session(token).mocambola
    if ROLE_MUCUA_PAJE not in mocambola.live_roles[balaio_smid]:
        raise HTTPException(
            status_code=401, detail="Mocambola não é um paje da mucua"
        )
    mucua = bbx.get_mucua_by_smid(
        balaio_smid=balaio_smid, mucua_smid=mucua_smid
    )
    if len(mucua.content) == 0:
        mucua.content.append(arquivo.filename)
    else:
        mucua.content[0] = arquivo.filename
    with (
        bbx.get_balaio_mucua_path(balaio_smid, mucua_smid) / mucua.content[0]
    ).open("wb") as arquivo_mucua:
        arquivo_mucua.write(arquivo.file.read())
        arquivo_mucua.close()

    bbx.put_mucua(balaio_smid, mucua_smid, mocambola, mucua)
    return {"detail": "success"}


@router.delete("/{balaio_slug_smid}/mucua/{mucua_slug_smid}")
async def del_mucua(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    mocambola: CurrentMocambola,
) -> Dict:
    """Deletar mucua."""
    balaio_smid, mucua_smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid, mucua_slug_smid=mucua_slug_smid
    )
    bbx.del_mucua(balaio_smid, mucua_smid, mocambola)
    return {"detail": "success"}
