#!/usr/bin/env python3

"""Grupo de rotas para autenticação."""
from pathlib import Path
from typing import Annotated, Dict, List, Optional

from fastapi import APIRouter, Depends, Header

from baobaxia.api.v2.endpoints.auth import (
    PermissionChecker,
    get_current_mocambola,
)
from baobaxia.root import Baobaxia, bbx
from baobaxia.saberes import Mocambola, Pedido

router = APIRouter(tags=["Pedidos"])

CurrentMocambola = Annotated[Mocambola, Depends(get_current_mocambola)]


@router.get("/{balaio_slug_smid}/{mucua_slug_smid}/pedidos")
async def list_pedidos(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    mocambola: CurrentMocambola,
) -> List[Pedido]:
    """Listar os Pedidos."""
    balaio_smid, mucua_smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid, mucua_slug_smid=mucua_slug_smid
    )
    return bbx.list_pedidos(balaio_smid, mucua_smid, mocambola)


@router.post("/{balaio_slug_smid}/{mucua_slug_smid}/pedido/{slug_smid:str}")
async def post_pedido(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    slug_smid: str,
    mocambola: CurrentMocambola,
    ttl: Optional[int] = 3,
) -> Dict:
    """Criar um pedido de copia de um saber"""
    balaio_smid, mucua_smid, smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid,
        mucua_slug_smid=mucua_slug_smid,
        slug_smid=slug_smid,
    )

    midia = bbx.get_midia(balaio_smid, mucua_smid, smid, mocambola)
    pedido = Pedido(
        name="Pedido  " + midia.titulo,
        path=Path("pedidos/"),
        midia_balaio_smid=midia.balaio_smid,
        midia_mucua_smid=midia.mucua_smid,
        midia_smid=midia.smid,
        ttl=ttl,
    )
    bbx.put_pedido(balaio_smid, mucua_smid, pedido, mocambola)
    midia.is_requested = True
    bbx.put_midia(balaio_smid, mucua_smid, midia, mocambola)
    return {"detail": "success"}


@router.post("/{balaio_slug_smid}/{mucua_slug_smid}/processar_pedidos")
async def processar_pedidos(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    token: str = Header(...),
) -> Dict:
    """Processar pedidos"""
    balaio_smid, mucua_smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid, mucua_slug_smid=mucua_slug_smid
    )
    mocambola = bbx.get_session(token).mocambola
    pedidos = bbx.list_pedidos(balaio_smid, mucua_smid, token)
    rotas = bbx.list_rotas(balaio_smid, mucua_smid, token)
    # Coletando tambem os pedidos das mucuas nas rotas
    for rota in rotas:
        for mucua_smid in rota.mucuas_smid:
            pedidos.extend(bbx.list_pedidos(balaio_smid, mucua_smid, token))
    bbx.get_copy(
        pedidos=pedidos,
        token=token,
        mocambola=mocambola,
        config=bbx.config,
    )
    return {"detail": "success"}


@router.delete("/{balaio_slug_smid}/{mucua_slug_smid}/pedido/{slug_smid:str}")
async def delete_pedido(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    slug_smid: str,
    mocambola: CurrentMocambola,
) -> Dict:
    """Remove um saber Pedido"""
    balaio_smid, mucua_smid, smid = bbx.extract_smid(
        balaio_slug_smid=balaio_slug_smid,
        mucua_slug_smid=mucua_slug_smid,
        slug_smid=slug_smid,
    )
    bbx.del_pedido(balaio_smid, mucua_smid, smid, mocambola)
    return {"detail": "success"}
