#!/usr/bin/env python3

"""Endpoints das Seleção."""
import re
import traceback
from pathlib import Path
from typing import Annotated, Dict, List, Optional

import cv2
from fastapi import (
    APIRouter,
    Body,
    Depends,
    File,
    Form,
    Header,
    HTTPException,
    Query,
    UploadFile,
)
from fastapi.responses import FileResponse
from pydantic import BaseModel

from baobaxia.api.v2.endpoints.auth import (
    PermissionChecker,
    get_current_mocambola,
)
from baobaxia.constants import ROLE_ACERVO_EDITOR, ROLE_ACERVO_PUBLISHER
from baobaxia.models.selecao import Selecao
from baobaxia.permissions.model_permissions import *
from baobaxia.root import Baobaxia, bbx
from baobaxia.saberes import Mocambola, Saber
from baobaxia.util import norm

router = APIRouter(tags=["Seleções"])

CurrentMocambola = Annotated[Mocambola, Depends(get_current_mocambola)]


@router.get("/selecao/find")
async def find_selecao(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: Optional[str] = Query(None),
    mucua_slug_smid: Optional[str] = Query(None),
    mocambola: Optional[Mocambola] = None,
    keywords: Optional[str] = Query(None),
    hashtags: Optional[str] = Query(None),
    status: Optional[str] = Query(None),
    creator: Optional[str] = Query(None),
    ordem_campo: Optional[str] = Query(None),
    ordem_decrescente: bool = Query(False),
    pag_tamanho: int = 12,
    pag_atual: int = 1,
) -> Dict:
    """Busca seleção de acordo com os parâmetros fornecidos"""
    if balaio_slug_smid is not None and balaio_slug_smid:
        balaio_smid = bbx.extract_smid(balaio_slug_smid)
    else:
        balaio_smid = None
    if mucua_slug_smid is not None and mucua_slug_smid:
        mucua_smid = bbx.extract_smid(mucua_slug_smid)
    else:
        mucua_smid = None
    kw_list = []
    if keywords is not None and len(keywords) > 0:
        kw_list = keywords.split(" ")
    ht_list = []
    if hashtags is not None and len(hashtags) > 0:
        # ht_list = hashtags.split(' ')
        ht_list = re.split("; |, ", hashtags)

    def filter_function(selecao):
        if status is not None and selecao.status != status:
            return False
        if creator is not None and selecao.creator != creator:
            return False
        match = True
        for kw in kw_list:
            if kw not in selecao.name:
                match = False
                break
        for ht in ht_list:
            if ht not in selecao.tags:
                match = False
                break
        return match

    def sorted_function(selecao):
        if ordem_campo is None:
            return 0
        elif hasattr(selecao, ordem_campo):
            return getattr(selecao, ordem_campo)
        else:
            return 0

    return bbx.find_selecaos(
        balaio_smid,
        mucua_smid,
        mocambola,
        filter_function=filter_function,
        sorted_function=sorted_function,
        sorted_reverse=ordem_decrescente,
        page_size=pag_tamanho,
        page_index=pag_atual,
    )


@router.get(
    "/{balaio_slug_smid}/{mucua_slug_smid}/selecao/{slug_smid}",
)
async def get_selecao(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    slug_smid: str,
) -> Selecao:
    """Retorna uma Seleção."""
    balaio_smid = bbx.extract_smid(balaio_slug_smid)
    mucua_smid = bbx.extract_smid(mucua_slug_smid)
    smid = bbx.extract_smid(slug_smid)
    return bbx.get_selecao(balaio_smid, mucua_smid, smid)


@router.post("/{balaio_slug_smid}/{mucua_slug_smid}/selecao")
@router.put("/{balaio_slug_smid}/{mucua_slug_smid}/selecao/{slug_smid:str}")
async def post_put_selecao(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    saber: Saber,
    mocambola: CurrentMocambola,
) -> Saber:
    """Cria ou atualiza uma Seleção."""
    balaio_smid = bbx.extract_smid(balaio_slug_smid)
    mucua_smid = bbx.extract_smid(mucua_slug_smid)
    return bbx.put_saber(balaio_smid, mucua_smid, saber, mocambola)


@router.delete("/{balaio_slug_smid}/{mucua_slug_smid}/selecao/{slug_smid}")
async def del_selecao(
    bbx: Annotated[Baobaxia, Depends(bbx)],
    balaio_slug_smid: str,
    mucua_slug_smid: str,
    slug_smid: str,
    mocambola: CurrentMocambola,
) -> Dict:
    """Remove uma Seleção."""
    balaio_smid = bbx.extract_smid(balaio_slug_smid)
    mucua_smid = bbx.extract_smid(mucua_slug_smid)
    smid = bbx.extract_smid(slug_smid)
    bbx.del_selecao(balaio_smid, mucua_smid, smid, mocambola)
    return {"detail": "success"}
