#!/usr/bin/env python3

import argparse
import os
from configparser import ConfigParser
from getpass import getpass
from pathlib import Path

from loguru import logger

from baobaxia.saberes import (
    Balaio,
    Mocambola,
    Mucua,
    SaberesConfig,
    SaberesDataStore,
)
from baobaxia.sankofa import Sankofa

parser = argparse.ArgumentParser("criar_mucua")
parser.add_argument(
    "--path",
    help="Caminho absoluto da pasta para os dados do Baobáxia",
    type=str,
)
parser.add_argument("--balaio", help="Nome do Balaio", type=str)
parser.add_argument(
    "--mucua", help="Nome da Mucua local onde instalar o Baobáxia", type=str
)
parser.add_argument(
    "--mocambola", help="Username para criar um Mocambola", type=str
)
parser.add_argument("--email", help="Email do Mocambola", type=str)
parser.add_argument("--password", help="Password para o Mocambola", type=str)
parser.add_argument(
    "--smid_len", help="Numero de carateres para os IDs", type=int
)
parser.add_argument(
    "--slug_name_len",
    help="Numero de carateres para os nomes abreviados",
    type=int,
)
parser.add_argument(
    "--slug_smid_len",
    help="Numero de carateres para os IDs abreviados",
    type=int,
)
parser.add_argument(
    "--default_ttl",
    help="Time To Live dos pedidos",
    type=int,
)


args = parser.parse_args()


def install_interactive(configure: bool = True):
    print("Instalador do Baobáxia")
    print("Entre com os dados do ambiente")
    configure = input("Deseja configurar a Mucua? [S/n] ")
    logger.debug(f"Configuração? {configure}")
    if configure in ["", "S", "s"]:
        path = input("Caminho do diretório de dados: ")
        balaio = input("Nome do balaio: ")
        mucua = input("Nome da mucua: ")
        print("Entre com os dados do mocambola")
        mocambola = input("Nome de usuário: ")
        email = input("E-mail: ")
        password = getpass("Senha: ")
        print("Entre com as configurações de chave primária")
        smid_len = input("Tamanho da chave aleatória (smid) ")
        slug_name_len = int(input("Tamanho do nome na chave primária: "))
        slug_smid_len = int(
            input("Tamanho da chave aleatória na chave primária: ")
        )
        default_ttl = int(input("Time To Live dos pedidos: "))
    else:
        path = "./tmp/instaler"
        balaio = "tmp_balaio"
        mucua = "tmp_mucua"
        mocambola = "tmp_mocambola"
        email = "mocambola@mucua.balaio"
        password = "tmp_password"
        smid_len = 7
        slug_name_len = 7
        slug_smid_len = 7
        default_ttl = 3

    install(
        path=path,
        balaio=balaio,
        mucua=mucua,
        mocambola=mocambola,
        email=email,
        password=password,
        smid_len=smid_len,
        slug_name_len=slug_name_len,
        slug_smid_len=slug_smid_len,
        default_ttl=default_ttl,
    )


def install(
    *,
    path: str,
    balaio: str,
    mucua: str,
    mocambola: str,
    email: str,
    password: str,
    smid_len: int,
    slug_name_len: int,
    slug_smid_len: int,
    default_ttl: int,
):
    """Instalador do Baobáxia

    :param path: Caminho absoluto da pasta para os dados do Baobáxia
    :type path: str
    :param balaio: Nome do Balaio
    :type balaio: str
    :param mucua: Nome da Mucua local onde instalar o Baobáxia
    :type mucua: str
    :param mocambola: Username para criar um Mocambola
    :type mocambola: str
    :param email: Email do Mocambola
    :type email: str
    :param password: Password para o Mocambola
    :type password: str
    :param smid_len: Numero de carateres para os IDs
    :type smid_len: int
    :param slug_name_len: Numero de carateres para os nomes abreviados
    :type slug_name_len: int
    :param slug_smid_len: Numero de carateres para os IDs abreviados
    :type slug_smid_len: int
    :param default_ttl: Time To Live dos pedidos
    :type default_ttl: int
    """

    default_data_path = Path(path)

    if default_data_path.is_absolute:
        config = SaberesConfig(
            default_data_path=default_data_path,
            default_balaio=balaio,
            default_mucua=mucua,
        )
        datastore = SaberesDataStore(config)

        balaio_dataset = datastore.create_balaio_dataset(mocambola=mocambola)
        balaio_saber = balaio_dataset.settle(
            Balaio(
                name=balaio,
                is_public=True,
                smid_len=smid_len,
                slug_name_len=slug_name_len,
                slug_smid_len=slug_smid_len,
                default_ttl=default_ttl,
            )
        )
        print("=========== BALAIO SABER ================")
        print(balaio_saber)
        mucua_dataset = datastore.create_mucua_dataset(
            balaio_smid=balaio_saber.smid, mocambola=mocambola
        )
        mucua_saber = mucua_dataset.settle(Mucua(name=mucua, is_public=True))
        print(mucua_saber)

        mocambola_dataset = datastore.create_dataset(
            Mocambola, balaio_saber.smid, mucua_saber.smid, mocambola
        )
        from .util import str_to_hash

        mocambola_saber = mocambola_dataset.settle(
            Mocambola(
                path=Path("mocambolas"),
                name=mocambola,
                username=mocambola,
                email=email,
                roles=["acervo.publisher", "acervo.editor", "mucua.paje"],
                password_hash=str_to_hash(password),
                is_public=True,
            )
        )

        #        mocambola_saber.path = (
        #            balaio_saber.path / mucua_saber.path / mocambola_saber.path
        #        )
        Sankofa.create_balaio(
            balaio=balaio_saber.slug,
            description=balaio_saber.slug,
            config=config,
        )

        #        mucua_saber.path = balaio_saber.slug / mucua_saber.path
        Sankofa.add(
            saberes=[mucua_saber, mocambola_saber],
            mocambola=mocambola_saber,
            config=config,
        )

    config_file = ConfigParser()

    config_file["default"] = {
        "default_data_path": path,
        "default_balaio": balaio_saber.smid,
        "default_mucua": mucua_saber.smid,
    }

    try:
        with open(
            os.path.join(os.path.expanduser("~"), ".baobaxia.conf"), "w"
        ) as writefile:
            config_file.write(writefile)
    except IOError:
        pass


def get_dummy_config(
    default_data_path: Path,
    *,
    default_balaio: str = "dummy_balaio",
    default_mucua: str = "dummy_mucua",
):
    return SaberesConfig(
        default_data_path=default_data_path,
        default_balaio=default_balaio,
        default_mucua=default_mucua,
    )


def get_dummy_datastore(
    default_data_path: Path,
    *,
    default_balaio: str = "dummy_balaio",
    default_mucua: str = "dummy_mucua",
):
    return SaberesDataStore(
        get_dummy_config(
            default_data_path,
            default_balaio=default_balaio,
            default_mucua=default_mucua,
        )
    )


def criar_mucua():
    install(**args.__dict__)
