#!/usr/bin/env python3

import os
import shutil
import sys
from configparser import ConfigParser
from math import ceil
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Any, List, Optional

from loguru import logger
from passlib.hash import bcrypt

from baobaxia.constants import SLUG_SEPARATOR
from baobaxia.models.artigo import Artigo
from baobaxia.models.midia import Midia, MidiaTipo
from baobaxia.models.selecao import Selecao
from baobaxia.saberes import (
    Balaio,
    Mocambola,
    Mucua,
    Pedido,
    Rota,
    Saber,
    SaberesConfig,
    SaberesDataStore,
)
from baobaxia.sankofa import Sankofa
from baobaxia.util import str_to_hash


class Baobaxia:

    def __init__(self, config: Optional[SaberesConfig] = None):
        logger.debug("!!! INICIALIZANDO BBX !!!")
        if config == None:
            config_file = ConfigParser()
            config_file.read(
                os.path.join(os.path.expanduser("~"), ".baobaxia.conf")
            )
            config = SaberesConfig.parse_obj(config_file["default"])
        self.config = config
        self.datastore = SaberesDataStore(self.config)
        self.reload_balaios()
        self.init_models()

    def init_models(self):
        logger.debug("Inicializando modelos.. ")
        # FIX: essa parte de base_path e criaçao de pastas tem que ser
        # colocada em outro lugar de inicializaçao do balaio

        base_path = (
            self.config.default_data_path
            / self.datastore.get_cached_path(
                smid=self.config.default_mucua,
                balaio_smid=self.config.default_balaio,
            )
        )

        # ACERVO - MIDIA
        pastas_por_tipo = {
            MidiaTipo.video: "videos",
            MidiaTipo.audio: "audios",
            MidiaTipo.imagem: "imagens",
            MidiaTipo.arquivo: "arquivos",
        }

        acervo_path = base_path / "acervo"
        if not acervo_path.exists():
            acervo_path.mkdir()
            for tipo, pasta in pastas_por_tipo.items():
                pasta_path = acervo_path / pasta
                if not pasta_path.exists():
                    pasta_path.mkdir()

        saberes_patterns_media = []
        for pattern in pastas_por_tipo.values():
            saberes_patterns_media.append("acervo/" + pattern + "/*/")

        self.discover_saberes(
            model=Midia, patterns=saberes_patterns_media, indexes_names=["tags"]
        )

        # BLOG - ARTIGO

        blog_path = base_path / "blog"
        if not blog_path.exists():
            blog_path.mkdir()

        saberes_patterns_artigo = ["blog/*/"]

        self.discover_saberes(
            model=Artigo,
            patterns=saberes_patterns_artigo,
            indexes_names=["tags"],
        )

        # SELECOES

        selecao_path = base_path / "selecao"
        if not selecao_path.exists():
            selecao_path.mkdir()

        saberes_patterns_selecao = ["selecao/*/"]

        self.discover_saberes(
            model=Selecao,
            patterns=saberes_patterns_selecao,
            indexes_names=["tags"],
        )

    def __call__(self):
        return self

    def reload_balaios(self):
        logger.debug(f"Baobáxia config: {self.config}")
        logger.debug(f"{self.datastore.config.default_data_path}")
        balaios_list = self.datastore.create_balaio_dataset().find_and_collect(
            "*/"
        )

        logger.debug(f"Balaios List: {balaios_list}")
        self._balaios = {}
        self._mucuas_por_balaio = {}
        self._mocambolas_por_mucua = {}
        self._roles_por_mocambolas = {}
        for balaio in balaios_list:
            logger.debug(f"Balaio: {balaio}")
            self._balaios[balaio.smid] = balaio
            self._mucuas_por_balaio[balaio.smid] = {}
            self._mocambolas_por_mucua[balaio.smid] = {}
            self._roles_por_mocambolas[balaio.smid] = {}
            mucuas = self.datastore.create_mucua_dataset(
                balaio.smid
            ).find_and_collect("*/")
            for mucua in mucuas:
                self._mucuas_por_balaio[balaio.smid][mucua.smid] = mucua
                logger.debug(f"Mucuas: {self._mucuas_por_balaio}")
                self._mocambolas_por_mucua[balaio.smid][mucua.smid] = {}
                mocambolas = self.datastore.create_dataset(
                    model=Mocambola,
                    balaio_smid=balaio.smid,
                    mucua_smid=mucua.smid,
                ).find_and_collect("mocambolas/*/")
                logger.debug(f"Mocambolas: {mocambolas}")
                for mocambola in mocambolas:
                    self._mocambolas_por_mucua[balaio.smid][mucua.smid][
                        mocambola.username
                    ] = mocambola
                    self._roles_por_mocambolas[balaio.smid][
                        mocambola.username
                    ] = {}
                    self._roles_por_mocambolas[balaio.smid][mocambola.username][
                        "role"
                    ] = mocambola.role
                    logger.debug(
                        f"Mocambola in cache: {self._mocambolas_por_mucua}"
                    )

        self.default_balaio = self._balaios[self.config.default_balaio]

        self.default_mucua = self._mucuas_por_balaio[
            self.config.default_balaio
        ][self.config.default_mucua]

    def extract_smid(
        self,
        balaio_slug_smid: Optional[str] = None,
        mucua_slug_smid: Optional[str] = None,
        slug_smid: Optional[str] = None,
    ):
        logger.debug(f"self._balaios: {self._balaios}")

        mucua_smid = None

        for bkey in self._balaios:
            logger.debug(f"bkey: {bkey}")
            if balaio_slug_smid and bkey.startswith(
                balaio_slug_smid.split(SLUG_SEPARATOR)[-1]
            ):
                balaio_smid = bkey
                if mucua_slug_smid:
                    for mkey in self._mucuas_por_balaio[bkey]:
                        if mkey.startswith(
                            mucua_slug_smid.split(SLUG_SEPARATOR)[-1]
                        ):
                            logger.debug(f"mucua_smid: {mkey}")
                            mucua_smid = mkey
                            if slug_smid:
                                for model in self.saberes[bkey][mkey]:
                                    for skey in self.saberes[bkey][mkey][model]:
                                        if skey.startswith(
                                            slug_smid.split(SLUG_SEPARATOR)[-1]
                                        ):
                                            smid = skey
                                            logger.debug(f"smid: {smid}")
                                            return [
                                                balaio_smid,
                                                mucua_smid,
                                                smid,
                                            ]
                    return balaio_smid, mucua_smid
                return balaio_smid
            if mucua_slug_smid:
                for mkey in self._mucuas_por_balaio[bkey]:
                    if mucua_slug_smid.split(SLUG_SEPARATOR)[-1] in mkey:
                        logger.debug(f"mucua_smid: {mkey}")
                        mucua_smid = mkey
        return mucua_smid

    def list_balaios(self):
        result = []
        for key, value in self._balaios.items():
            if value.is_public:
                if value not in result:
                    result.append(value.copy())
        return result

    def get_balaio(self, smid: str):
        if self._balaios[smid].is_public:
            return self._balaios[smid].copy()
        return None

    def get_default_balaio(self):
        return self.default_balaio

    def post_balaio(
        self, balaio: Balaio, default_mucua_name: str, mocambola: Mocambola
    ):
        if balaio.default_mucua:
            balaio.default_mucua = self.config.default_mucua
        dataset = self.datastore.create_balaio_dataset(
            mocambola=mocambola.username
        )
        balaio = dataset.settle(balaio)
        self._balaios[balaio.smid] = balaio
        self._mucuas_por_balaio[balaio.smid] = {}
        self._mocambolas_por_mucua[balaio.smid] = {}
        Sankofa.create_balaio(
            balaio=balaio.slug, description=balaio.name, config=self.config
        )
        if balaio.default_mucua is not None:
            mucua_existe = self.get_mucua_by_smid(
                mucua_smid=balaio.default_mucua
            )
            if mucua_existe:
                mucua_name = mucua_existe.name
            else:
                mucua_name = default_mucua_name
            mucua = self.put_mucua(
                balaio_smid=balaio.smid,
                name=mucua_name,
                mucua_smid=balaio.default_mucua,
                mocambola=mocambola,
                is_public=balaio.is_public,  # se o balaio is_public a mucua tambem
            )
            dataset.settle(balaio)
        return balaio.copy()

    def put_balaio(self, *, balaio: Balaio, mocambola: Mocambola):
        dataset = self.datastore.create_balaio_dataset(
            mocambola=mocambola.username
        )
        if not balaio.default_mucua:
            balaio.default_mucua = self.config.default_mucua
        balaio = dataset.settle(balaio)
        self._balaios[balaio.smid] = balaio
        return balaio.copy()

    def del_balaio(self, balaio_smid: str):
        if balaio_smid == self.default_balaio.smid:
            # Baobaxia não pode deixar para trás o balaio de seus próprios saberes.
            raise RuntimeError("Cannot delete balaio padrão.")
        dataset = self.datastore.create_balaio_dataset()
        balaio = self._balaios[balaio_smid]
        Sankofa.destroy_balaio(balaio.slug, config=self.config)
        dataset.drop(balaio.smid)

        del self._balaios[balaio_smid]
        del self._mucuas_por_balaio[balaio_smid]
        del self._mocambolas_por_mucua[balaio_smid]

    def list_mucuas(self, balaio_smid: Optional[str] = None):
        result = []
        balaios = [
            b for b in self._balaios if not balaio_smid or b == balaio_smid
        ]
        for b_smid in balaios:
            for key, value in self._mucuas_por_balaio[b_smid].items():
                if value.is_public:
                    if value.smid not in [result.smid for result in result]:
                        result.append(value)
        return result

    def get_mucua(self, balaio_smid: str, mucua_smid: str):
        if self._mucuas_por_balaio[balaio_smid][mucua_smid].is_public:
            return self._mucuas_por_balaio[balaio_smid][mucua_smid].copy()
        return None

    def get_mucua_by_smid(
        self, mucua_smid: str, balaio_smid: Optional[str] = None
    ):
        if balaio_smid is not None:
            if self._mucuas_por_balaio[balaio_smid].get(mucua_smid):
                return self._mucuas_por_balaio[balaio_smid][mucua_smid]
        else:
            for b_smid in self._mucuas_por_balaio:
                if self._mucuas_por_balaio[b_smid].get(mucua_smid):
                    return self._mucuas_por_balaio[b_smid][mucua_smid]
                return None

    def get_default_mucua(self):
        return self.default_mucua

    def _atualiza_mucua_em_balaios(self, mucua: Mucua, mocambola: Mocambola):
        mucua_path = self.get_balaio_mucua_path(mucua.balaio_smid, mucua.smid)
        for balaio_smid in self._mucuas_por_balaio:
            if balaio_smid != mucua.balaio_smid:
                if self._mucuas_por_balaio[balaio_smid].get(mucua.smid):
                    dataset = self.datastore.create_mucua_dataset(
                        balaio_smid=balaio_smid, mocambola=mocambola
                    )
                    mucua.balaio_smid = balaio_smid
                    mucua = dataset.settle(mucua)
                    for content in mucua.content:
                        path_fonte = mucua_path / content
                        path_destino = (
                            self.get_balaio_mucua_path(
                                balaio_smid=mucua.balaio_smid,
                                mucua_smid=mucua.smid,
                            )
                            / content["filename"]
                        )
                        shutil.copy(path_fonte, path_destino)
                    Sankofa.add(
                        saberes=[mucua],
                        mocambola=mocambola,
                        config=self.config,
                    )

    def post_mucua(
        self,
        *,
        balaio_smid: str,
        mucua: Optional[Mucua] = None,
        is_public: Optional[bool] = True,
        mocambola: Mocambola,
    ):
        dataset = self.datastore.create_mucua_dataset(
            balaio_smid=balaio_smid, mocambola=mocambola.username
        )
        mucua = dataset.settle(Mucua(name=mucua.name, is_public=is_public))
        self._mucuas_por_balaio[balaio_smid][mucua.smid] = mucua
        self._atualiza_mucua_em_balaios(mucua=mucua, mocambola=mocambola)
        Sankofa.add(saberes=[mucua], mocambola=mocambola, config=self.config)
        return mucua.copy()

    def put_mucua(
        self,
        *,
        balaio_smid: str,
        name: Optional[str] = None,
        mucua_smid: Optional[str] = None,
        mucua: Optional[Mucua] = None,
        is_public: Optional[bool] = True,
        mocambola: Mocambola,
    ):
        dataset = self.datastore.create_mucua_dataset(
            balaio_smid=balaio_smid, mocambola=mocambola.username
        )
        if mucua is not None:
            if mucua.smid != self.default_mucua.smid:
                raise RuntimeError("Não é possivel alterar outra mucua")
            mucua = dataset.settle(mucua)
        elif mucua_smid is None:
            mucua = dataset.settle(Mucua(name=name, is_public=is_public))
        elif mucua_smid is not None and not self._mucuas_por_balaio[
            balaio_smid
        ].get(mucua_smid):
            if mucua_smid != self.default_mucua.smid:
                raise RuntimeError("Não é possivel alterar outra mucua")
            mucua = dataset.settle(
                Mucua(
                    name=name,
                    balaio_smid=balaio_smid,
                    smid=mucua_smid,
                    is_public=is_public,
                )
            )
            self._mucuas_por_balaio[balaio_smid][mucua_smid] = mucua
        else:
            mucua = self._mucuas_por_balaio[balaio_smid][mucua_smid]
            if name:
                mucua.name = name
            mucua = dataset.settle(mucua)
        self._atualiza_mucua_em_balaios(mucua=mucua, mocambola=mocambola)
        Sankofa.add(saberes=[mucua], mocambola=mocambola, config=self.config)
        return mucua.copy()

    def del_mucua(
        self, balaio_smid: str, mucua_smid: str, mocambola: Mocambola
    ):
        if (
            balaio_smid == self.default_balaio.smid
            and mucua_smid == self.default_balaio.mucua_smid
        ):
            # Baobaxia não pode deixar para trás a mucua de seus próprios saberes.
            raise RuntimeError("Cannot delete mucua padrão.")

        mucua = self._mucuas_por_balaio[balaio_smid][mucua_smid]

        dataset = self.datastore.create_mucua_dataset(
            balaio_smid=balaio_smid,
        )
        Sankofa.remove(
            saberes=[mucua],
            mocambola=mocambola,
            config=self.config,
        )
        dataset.drop(mucua)
        del self._mucuas_por_balaio[balaio_smid][mucua_smid]
        del self._mocambolas_por_mucua[balaio_smid][mucua_smid]

    def list_mocambolas(
        self,
        balaio_smid: Optional[str] = None,
        mucua_smid: Optional[str] = None,
    ):
        result = []

        if balaio_smid:
            if not self._mucuas_por_balaio.get(balaio_smid):
                raise RuntimeError("Balaio doesn't exists.")

            balaios = [balaio_smid]

            if mucua_smid:
                mucuas = [
                    m
                    for m in self._mucuas_por_balaio[balaio_smid]
                    if m == mucua_smid
                ]
            mucuas = [m for m in self._mucuas_por_balaio[balaio_smid]]
        else:
            if mucua_smid:
                balaios = self._balaios
                mucuas = []
                for b in self._balaios:
                    for m in self._mucuas_por_balaio[b]:
                        if m == mucua_smid:
                            mucuas.append(self._mucuas_por_balaio[b][m].smid)
            else:
                balaios = self._balaios
                mucuas = []
                for b in self._balaios:
                    for m in self._mucuas_por_balaio[b]:
                        mucuas.append(self._mucuas_por_balaio[b][m].smid)

        for balaio_smid in balaios:
            for mucua in mucuas:
                logger.debug(
                    f"LM - MUCUAS: {mucuas}, MOCAMBOLAS: {self._mocambolas_por_mucua}"
                )
                for username in self._mocambolas_por_mucua[balaio_smid][mucua]:
                    result.append(
                        self._mocambolas_por_mucua[balaio_smid][mucua][
                            username
                        ].copy(
                            exclude={
                                "password_hash",
                                "recovery_answer_hash",
                            }
                        )
                    )
        return result

    def get_mocambola(
        self,
        username: str,
        balaio_smid: Optional[str] = None,
        mucua_smid: Optional[str] = None,
    ):
        logger.debug(
            f"GET_MOCAMBOLA, username: {username}, balaio: {balaio_smid}, mucua: {mucua_smid}"
        )
        if balaio_smid is None:
            balaio_smid = self.default_balaio.smid
        if mucua_smid is None:
            mucua_smid = self.default_mucua.smid
        if self._mocambolas_por_mucua[balaio_smid][mucua_smid].get(username):
            if self._mocambolas_por_mucua[balaio_smid][mucua_smid][
                username
            ].is_public:
                return self._mocambolas_por_mucua[balaio_smid][mucua_smid][
                    username
                ].copy(exclude={"password_hash", "recovery_answer_hash"})
            return None
        return None

    def get_full_mocambola(
        self,
        username: str,
        balaio_smid: Optional[str] = None,
        mucua_smid: Optional[str] = None,
    ):
        if balaio_smid is None:
            balaio_smid = self.default_balaio.smid
        if mucua_smid is None:
            mucua_smid = self.default_mucua.smid
        logger.debug(f"Mocambolas/Mucua: {self._mocambolas_por_mucua}")
        if self._mocambolas_por_mucua[balaio_smid][mucua_smid][
            username
        ].is_public:
            logger.debug("Mocambola is public")
            return self._mocambolas_por_mucua[balaio_smid][mucua_smid][
                username
            ].copy()
        return None

    def put_mocambola(
        self,
        balaio_smid: str,
        mucua_smid: str,
        mocambola: Mocambola,
        current_mocambola: Optional[Mocambola] = None,
        recovery_answer: Optional[str] = None,
    ):
        dataset = self.datastore.create_dataset(
            model=Mocambola,
            balaio_smid=balaio_smid,
            mucua_smid=mucua_smid,
            mocambola=mocambola.username,
        )
        mocambola = dataset.settle(mocambola)
        self._mocambolas_por_mucua[balaio_smid][mucua_smid][
            mocambola.username
        ] = mocambola

        if current_mocambola:
            creator = current_mocambola
        else:
            creator = mocambola

        Sankofa.add(saberes=[mocambola], mocambola=creator, config=self.config)

        return mocambola.copy(exclude={"password_hash", "recovery_answer_hash"})

    def del_mocambola(
        self,
        balaio_smid: str,
        mucua_smid: str,
        username: str,
        mocambola: Optional[Mocambola],
    ):
        if mocambola is None:
            mocambola = self._mocambolas_por_mucua[balaio_smid][mucua_smid][
                username
            ]
        self.datastore.create_dataset(
            model=Mocambola,
            balaio_smid=balaio_smid,
            mucua_smid=mucua_smid,
        ).drop(mocambola.path)
        del self._mocambolas_por_mucua[balaio_smid][mucua_smid][username]
        Sankofa.remove(
            saberes=[mocambola],
            mocambola=mocambola,
            config=self.config,
        )

    def add_mocambola_balaio(
        self,
        balaio_smid: str,
        mucua_smid: str,
        novo_balaio_smid: str,
        username: str,
        mocambola: Optional[Mocambola],
    ):
        if mocambola is None:
            mocambola = self._mocambolas_por_mucua[balaio_smid][mucua_smid][
                username
            ]
        mocambola.balaio_smid = novo_balaio_smid
        dataset = self.datastore.create_dataset(
            model=Mocambola,
            balaio_smid=novo_balaio_smid,
            mucua_smid=mucua_smid,
            mocambola=mocambola,
        )
        mocambola = dataset.settle(mocambola)
        Sankofa.add(
            saberes=[mocambola], mocambola=mocambola, config=self.config
        )
        self._mocambolas_por_mucua[novo_balaio_smid][mucua_smid][
            username
        ] = mocambola
        return True

    def _add_to_index(self, balaio: str, mucua: str, field: str, saber: Saber):
        if field not in self.indexes_names:
            return
        if balaio not in self.indexes:
            self.indexes[balaio] = {}
        if mucua not in self.indexes[balaio]:
            self.indexes[balaio][mucua] = {}
        if field not in self.indexes[balaio][mucua]:
            self.indexes[balaio][mucua][field] = {}
        for idx in self.indexes_names[field]:
            if idx not in self.indexes[balaio][mucua][field]:
                self.indexes[balaio][mucua][field][idx] = {}
            val = getattr(saber, idx)
            if isinstance(val, list):
                for val_item in val:
                    logger.debug(f"VAL_ITEM: {val_item}")
                    if isinstance(val_item, dict):
                        for key, value in val_item.items():
                            self.indexes[balaio][mucua][field][idx][key] = {}
                            logger.debug(
                                f"Add_to_index, INDEX: {self.indexes[balaio][mucua][field][idx]}"
                            )
                            self.indexes[balaio][mucua][field][idx][key] = value
                    else:
                        if (
                            val_item
                            not in self.indexes[balaio][mucua][field][idx]
                        ):
                            self.indexes[balaio][mucua][field][idx][
                                val_item
                            ] = set()
                            self.indexes[balaio][mucua][field][idx][
                                val_item
                            ].add(saber.smid)
            else:
                if val not in self.indexes[balaio][mucua][field][idx]:
                    self.indexes[balaio][mucua][field][idx][val] = set()
                self.indexes[balaio][mucua][field][idx][val].add(saber.smid)

    def _remove_from_index(
        self, balaio: str, mucua: str, field: str, saber: Saber
    ):
        if field not in self.indexes_names:
            return
        if balaio not in self.indexes:
            return
        if mucua not in self.indexes[balaio]:
            return
        if field not in self.indexes[balaio][mucua]:
            return
        for idx in self.indexes_names[field]:
            if idx not in self.indexes[balaio][mucua][field]:
                continue
            val = getattr(saber, idx)
            if isinstance(val, list):
                for val_item in val:
                    if isinstance(val_item, dict):
                        for key, value in val_item.items():
                            if self.indexes[balaio][mucua][field][idx][key]:
                                continue
                            self.indexes[balaio][mucua][field][idx][
                                key
                            ].discard(value)
                    else:
                        if (
                            val_item
                            not in self.indexes[balaio][mucua][field][idx]
                        ):
                            continue
                        self.indexes[balaio][mucua][field][idx][
                            val_item
                        ].discard(saber.smid)
            else:
                if val not in self.indexes[balaio][mucua][field][idx]:
                    continue
                self.indexes[balaio][mucua][field][idx][val].discard(saber.smid)

    def get_balaio_mucua_path(
        self, balaio_smid, mucua_smid=None, relative=False
    ):
        datapath = Path(self.config.default_data_path)
        if mucua_smid is not None:
            balaiomucuapath = Path(
                self.datastore.get_cached_path(
                    smid=mucua_smid, balaio_smid=balaio_smid
                )
            )
        else:
            balaiomucuapath = Path(
                self.datastore.get_cached_path(smid=balaio_smid)
            )
        if relative:
            return balaiomucuapath
        else:
            return datapath / balaiomucuapath

    def get_path_by_index(
        self,
        balaio_smid: str,
        mucua_smid: str,
        field: str,
        index: str,
        value: Any,
    ):
        result = []
        if (
            balaio_smid not in self.indexes
            and mucua_smid in self.indexes[balaio_smid]
            and field in self.indexes[balaio_smid][mucua_smid]
            and index in self.indexes[balaio_smid][mucua_smid][field]
            and value in self.indexes[balaio_smid][mucua_smid][field][index]
        ):
            # FIX o que seria expand?
            result.extend(
                self.indexes[balaio_smid][mucua_smid][field][index][value]
            )
        return result

    def _check_cache(
        self,
        model: type,
        field: str,
        patterns: List[str],
        balaio_smid: str,
        mucua_smid: str,
        update: bool = False,
    ):
        if (
            not update
            and balaio_smid in self.saberes
            and mucua_smid in self.saberes[balaio_smid]
            and field in self.saberes[balaio_smid][mucua_smid]
        ):
            return
        if balaio_smid not in self.saberes:
            self.saberes[balaio_smid] = {}
        if mucua_smid not in self.saberes[balaio_smid]:
            self.saberes[balaio_smid][mucua_smid] = {}
        self.saberes[balaio_smid][mucua_smid][field] = {}

        if balaio_smid not in self.indexes:
            self.indexes[balaio_smid] = {}
        if mucua_smid not in self.indexes[balaio_smid]:
            self.indexes[balaio_smid][mucua_smid] = {}
        self.indexes[balaio_smid][mucua_smid][field] = {}
        for idx_name in self.indexes_names:
            self.indexes[balaio_smid][mucua_smid][field][idx_name] = {}

        dataset = self.datastore.create_dataset(
            model=model, balaio_smid=balaio_smid, mucua_smid=mucua_smid
        )

        logger.debug(f"BASEPATH: {dataset.get_base_path()}")
        for pattern in patterns:
            saberes = dataset.find_and_collect(pattern)
            for saber in saberes:
                self.saberes[balaio_smid][mucua_smid][field][saber.smid] = saber
                self._add_to_index(balaio_smid, mucua_smid, field, saber)
        logger.debug(f"INDEX: {self.indexes}")

    def discover_saberes(
        self,
        *,
        model: type,
        patterns: List[str],
        field_name: Optional[str] = None,
        list_field_name: Optional[str] = None,
        indexes_names: List[str] = [],
    ):
        if field_name is None:
            field_name = model.__name__.lower()
        if list_field_name is None:
            list_field_name = field_name + "s"

        if not hasattr(self, "saberes"):
            self.saberes = {}
        if not hasattr(self, "indexes"):
            self.indexes = {}
        if not hasattr(self, "indexes_names"):
            self.indexes_names = {}
        self.indexes_names[field_name] = indexes_names

        def list_method_template(
            balaio_smid: Optional[str], mucua_smid: Optional[str]
        ):
            result = []
            if not balaio_smid:
                for b_smid in self._balaios:
                    if not mucua_smid:
                        for m_smid in self._mucuas_por_balaio[b_smid]:
                            self._check_cache(
                                model, field_name, patterns, b_smid, m_smid
                            )
                            for key, saber in self.saberes[b_smid][m_smid][
                                field_name
                            ].items():
                                if saber.is_public:
                                    result.append(saber.copy())
                                    return result
                    else:
                        self._check_cache(
                            model, field_name, patterns, b_smid, m_smid
                        )
                        for key, saber in self.saberes[b_smid][m_smid][
                            field_name
                        ].items():
                            if saber.is_public:
                                result.append(saber.copy())
                                return result
            else:
                if not mucua_smid:
                    for m_smid in self._mucuas_por_balaio[balaio_smid]:
                        self._check_cache(
                            model, field_name, patterns, balaio_smid, m_smid
                        )
                        for key, saber in self.saberes[balaio_smid][m_smid][
                            field_name
                        ].items():
                            if saber.is_public:
                                result.append(saber.copy())
                                return result
                else:
                    self._check_cache(
                        model, field_name, patterns, balaio_smid, m_smid
                    )
                    for key, saber in self.saberes[balaio_smid][m_smid][
                        field_name
                    ].items():
                        if saber.is_public:
                            result.append(saber.copy())
                            return result

        setattr(self, "list_" + list_field_name, list_method_template)

        def find_method_template(
            balaio_smid: Optional[str] = None,
            mucua_smid: Optional[str] = None,
            mocambola: Optional[Mocambola] = None,
            filter_function: Optional[callable] = None,
            sorted_function: Optional[callable] = None,
            sorted_reverse: bool = False,
            page_size: int = 0,
            page_index: int = 0,
        ):
            result = []
            if balaio_smid is not None and mucua_smid is not None:
                for key, saber in self.saberes[balaio_smid][mucua_smid][
                    field_name
                ].items():
                    if saber.is_public:
                        if filter_function is None or filter_function(saber):
                            result.append(saber.copy())
            elif balaio_smid is not None:
                for m_smid in self._mucuas_por_balaio[balaio_smid]:
                    self._check_cache(
                        model, field_name, patterns, balaio_smid, m_smid
                    )
                    for key, saber in self.saberes[balaio_smid][m_smid][
                        field_name
                    ].items():
                        if saber.is_public:
                            if filter_function is None or filter_function(
                                saber
                            ):
                                result.append(saber.copy())
            elif mucua_smid is not None:
                for b_smid in self._balaios:
                    if mucua_smid in self._mucuas_por_balaio[b_smid]:
                        self._check_cache(
                            model, field_name, patterns, b_smid, mucua_smid
                        )
                        for key, saber in self.saberes[b_smid][mucua_smid][
                            field_name
                        ].items():
                            if saber.is_public:
                                if filter_function is None or filter_function(
                                    saber
                                ):
                                    result.append(saber.copy())
            else:
                for b_smid in self._balaios:
                    for m_smid in self._mucuas_por_balaio[b_smid]:
                        self._check_cache(
                            model, field_name, patterns, b_smid, m_smid
                        )
                        for key, saber in self.saberes[b_smid][m_smid][
                            field_name
                        ].items():
                            if saber.is_public:
                                if filter_function is None or filter_function(
                                    saber
                                ):
                                    result.append(saber.copy())
            if sorted_function is not None:
                result = sorted(
                    result, key=sorted_function, reverse=sorted_reverse
                )
            item_count = len(result)
            page_count = ceil(item_count / page_size)
            first_item = (page_index - 1) * page_size
            last_item = first_item + page_size
            return {
                "item_count": item_count,
                "page_count": page_count,
                "items": result[first_item:last_item],
            }

        setattr(self, "find_" + list_field_name, find_method_template)

        def get_method_template(balaio_smid: str, mucua_smid: str, smid: str):
            self._check_cache(
                model, field_name, patterns, balaio_smid, mucua_smid
            )
            if smid in self.saberes[balaio_smid][mucua_smid][field_name]:
                if self.saberes[balaio_smid][mucua_smid][field_name][
                    smid
                ].is_public:
                    return self.saberes[balaio_smid][mucua_smid][field_name][
                        smid
                    ].copy()
            return None

        setattr(self, "get_" + field_name, get_method_template)

        def put_method_template(
            balaio_smid: str, mucua_smid: str, saber: model, mocambola=Mocambola
        ):
            logger.info("Inside put template!!!")
            self._check_cache(
                model, field_name, patterns, balaio_smid, mucua_smid
            )
            saber = self.datastore.create_dataset(
                model=model,
                balaio_smid=balaio_smid,
                mucua_smid=mucua_smid,
                mocambola=mocambola,
            ).settle(saber)
            if saber.smid in self.saberes[balaio_smid][mucua_smid][field_name]:
                self._remove_from_index(
                    balaio_smid,
                    mucua_smid,
                    field_name,
                    self.saberes[balaio_smid][mucua_smid][field_name][
                        saber.smid
                    ],
                )
            self.saberes[balaio_smid][mucua_smid][field_name][
                saber.smid
            ] = saber
            self._add_to_index(balaio_smid, mucua_smid, field_name, saber)
            Sankofa.add(
                saberes=[saber], mocambola=mocambola, config=self.config
            )
            return saber.copy()

        setattr(self, "put_" + field_name, put_method_template)

        def del_method_template(
            balaio_smid: str, mucua_smid: str, smid: str, mocambola: Mocambola
        ):
            self._check_cache(
                model, field_name, patterns, balaio_smid, mucua_smid
            )
            dataset = self.datastore.create_dataset(
                model=model,
                balaio_smid=balaio_smid,
                mucua_smid=mucua_smid,
                mocambola=mocambola,
            )
            saber = self.saberes[balaio_smid][mucua_smid][field_name][smid]
            Sankofa.remove(
                saberes=[saber], mocambola=mocambola, config=self.config
            )
            del self.saberes[balaio_smid][mucua_smid][field_name][smid]
            self._remove_from_index(balaio_smid, mucua_smid, field_name, saber)
            dataset.drop(saber.path)

        setattr(self, "del_" + field_name, del_method_template)

    def set_password(
        self,
        *,
        balaio_smid: str,
        mucua_smid: str,
        new_password: str,
        recovery_answer: Optional[str] = None,
        mocambola: Mocambola,
    ):
        mocambola.password_hash = str_to_hash(new_password)
        if recovery_answer:
            mocambola.recovery_answer_hash = str_to_hash(recovery_answer)
        dataset = self.datastore.create_dataset(
            Mocambola, mocambola.balaio_smid, mocambola.mucua_smid, mocambola
        )
        mocambola = dataset.settle(mocambola)
        self._mocambolas_por_mucua[balaio_smid][mucua_smid][
            mocambola.username
        ] = mocambola

        Sankofa.add(
            saberes=[mocambola], mocambola=mocambola, config=self.config
        )
        return {"detail": "success"}

        return {"detail": "success"}

    def set_recovery(
        self,
        *,
        new_recovery_question: str,
        new_recovery_answer: str,
        recovery_answer: Optional[str] = None,
        mocambola: Mocambola,
    ):
        # FIX: Check recovery_answer
        mocambola.recovery_question = new_recovery_question
        mocambola.recovery_answer_hash = str_to_hash(new_recovery_answer)
        dataset = self.datastore.create_dataset(
            Mocambola, mocambola.balaio_smid, mocambola.mucua_smid, mocambola
        )
        mocambola = dataset.settle(mocambola)
        Sankofa.add(
            saberes=[mocambola], mocambola=mocambola, config=self.config
        )

    def list_rotas(self, balaio_smid: str, mucua_smid: str):
        model = Rota
        field_name = "rota"
        patterns = ["rotas/*/"]
        self._check_cache(model, field_name, patterns, balaio_smid, mucua_smid)
        result = []
        for key, saber in self.saberes[balaio_smid][mucua_smid][
            field_name
        ].items():
            if saber.is_public:
                result.append(saber.copy())
        return result

    def get_rota(self, balaio_smid: str, mucua_smid: str, smid: str):
        model = Rota
        field_name = "rota"
        patterns = ["rotas/*/"]
        self._check_cache(model, field_name, patterns, balaio_smid, mucua_smid)
        if smid in self.saberes[balaio_smid][mucua_smid][field_name]:
            if self.saberes[balaio_smid][mucua_smid][field_name][
                smid
            ].is_public:
                return self.saberes[balaio_smid][mucua_smid][field_name][
                    smid
                ].copy()
            return None

    def put_rota(
        self,
        balaio_smid: str,
        mucua_smid: str,
        saber: Rota,
        mocambola: Mocambola,
    ):
        model = Rota
        field_name = "rota"
        patterns = ["rotas/*/"]
        if saber.smid is None:
            saber.path = Path("rotas")
        saber = self.datastore.create_dataset(
            model=model,
            balaio_smid=balaio_smid,
            mucua_smid=mucua_smid,
            mocambola=mocambola,
        ).settle(saber)
        self._check_cache(model, field_name, patterns, balaio_smid, mucua_smid)
        if saber.smid in self.saberes[balaio_smid][mucua_smid][field_name]:
            self._remove_from_index(
                balaio_smid,
                mucua_smid,
                field_name,
                self.saberes[balaio_smid][mucua_smid][field_name][saber.smid],
            )
        self.saberes[balaio_smid][mucua_smid][field_name][saber.smid] = saber
        self._add_to_index(balaio_smid, mucua_smid, field_name, saber)

        Sankofa.add(saberes=[saber], mocambola=mocambola, config=self.config)
        return saber.copy()

    def del_rota(
        self, balaio_smid: str, mucua_smid: str, smid: str, mocambola: Mocambola
    ):
        model = Rota
        field_name = "rota"
        patterns = ["rotas/*/"]
        self._check_cache(model, field_name, patterns, balaio_smid, mucua_smid)
        dataset = self.datastore.create_dataset(
            model=model,
            balaio_smid=balaio_smid,
            mucua_smid=mucua_smid,
            mocambola=mocambola,
        )
        saber = self.saberes[balaio_smid][mucua_smid][field_name][smid]
        Sankofa.remove(saberes=[saber], mocambola=mocambola, config=self.config)
        del self.saberes[balaio_smid][mucua_smid][field_name][smid]
        self._remove_from_index(balaio_smid, mucua_smid, field_name, saber)
        dataset.drop(saber.path)

    def atualizar_rotas(self, rotas, config):
        for rota in rotas:
            if rota.is_active:
                for mucua_smid in rota.mucuas_smid:
                    mucua = self.get_mucua_by_smid(
                        balaio_smid=rota.balaio_smid, mucua_smid=mucua_smid
                    )

                    for uri in [
                        uri for uri in dir(mucua) if uri.startswith("uri_")
                    ]:
                        uri_value = getattr(mucua, uri)
                        if uri_value is not None:
                            Sankofa.set_remote(
                                balaio_path=self.get_balaio_mucua_path(
                                    rota.balaio_smid, relative=True
                                ),
                                name=mucua_smid + "_" + uri,
                                uri=uri_value,
                                config=config,
                            )

    def list_pedidos(self, balaio_smid: str, mucua_smid: str):
        model = Pedido
        field_name = "pedido"
        patterns = ["pedidos/*/"]
        self._check_cache(model, field_name, patterns, balaio_smid, mucua_smid)
        result = []
        for key, saber in self.saberes[balaio_smid][mucua_smid][
            field_name
        ].items():
            if saber.is_public:
                result.append(saber.copy())
        return result

    def get_pedido(self, balaio_smid: str, mucua_smid: str, smid: str):
        model = Pedido
        field_name = "pedido"
        patterns = ["pedidos/*/"]
        self._check_cache(model, field_name, patterns, balaio_smid, mucua_smid)
        if smid in self.saberes[balaio_smid][mucua_smid][field_name]:
            if self.saberes[balaio_smid][mucua_smid][field_name][
                smid
            ].is_public:
                return self.saberes[balaio_smid][mucua_smid][field_name][
                    smid
                ].copy()
            return None

    def put_pedido(
        self,
        balaio_smid: str,
        mucua_smid: str,
        saber: Pedido,
        mocambola: Mocambola,
    ):
        model = Pedido
        field_name = "pedido"
        patterns = ["pedidos/*/"]
        self._check_cache(model, field_name, patterns, balaio_smid, mucua_smid)
        if saber.smid is None:
            saber.path = Path("pedidos")
        saber = self.datastore.create_dataset(
            model=model,
            balaio_smid=balaio_smid,
            mucua_smid=mucua_smid,
            mocambola=mocambola,
        ).settle(saber)
        if saber.smid in self.saberes[balaio_smid][mucua_smid][field_name]:
            self._remove_from_index(
                balaio_smid,
                mucua_smid,
                field_name,
                self.saberes[balaio_smid][mucua_smid][field_name][saber.smid],
            )
        self.saberes[balaio_smid][mucua_smid][field_name][saber.smid] = saber
        self._add_to_index(balaio_smid, mucua_smid, field_name, saber)
        Sankofa.add(saberes=[saber], mocambola=mocambola, config=self.config)
        return saber.copy()

    def del_pedido(
        self, balaio_smid: str, mucua_smid: str, smid: str, mocambola: Mocambola
    ):
        model = Pedido
        field_name = "pedido"
        patterns = ["pedidos/*/"]
        self._check_cache(model, field_name, patterns, balaio_smid, mucua_smid)
        dataset = self.datastore.create_dataset(
            model=model,
            balaio_smid=balaio_smid,
            mucua_smid=mucua_smid,
            mocambola=mocambola,
        )
        saber = self.saberes[balaio_smid][mucua_smid][field_name][smid]
        Sankofa.remove(saberes=[saber], mocambola=mocambola, config=self.config)
        del self.saberes[balaio_smid][mucua_smid][field_name][smid]
        self._remove_from_index(balaio_smid, mucua_smid, field_name, saber)
        dataset.drop(saber.path)

    def get_copy(self, pedidos, mocambola, config):
        for pedido in pedidos:
            midia_solicitada = self.get_midia(
                pedido.midia_balaio_smid,
                pedido.midia_mucua_smid,
                pedido.midia_smid,
            )
            midia_solicitada.path = (
                self.get_balaio_mucua_path(
                    balaio_smid=pedido.midia_balaio_smid,
                    mucua_smid=pedido.midia_mucua_smid,
                    relative=True,
                )
                / midia_solicitada.path
            )
            Sankofa.get(saberes=[midia_solicitada], config=config)
            pedido.ttl -= 1
            if pedido.ttl <= 0:
                pedido.path = (
                    self.get_balaio_mucua_path(
                        balaio_smid=pedido.midia_balaio_smid,
                        mucua_smid=pedido.midia_mucua_smid,
                        relative=True,
                    )
                    / pedido.path
                )
                Sankofa.remove(
                    saberes=[pedido], mocambola=mocambola, config=config
                )
                self.del_pedido(
                    pedido.balaio_smid,
                    pedido.mucua_smid,
                    pedido.smid,
                    mocambola,
                )
            else:
                self.put_pedido(
                    pedido.balaio_smid, pedido.mucua_smid, pedido, mocambola
                )
                if not midia_solicitada.is_requested:
                    Sankofa.remove(
                        saberes=[midia_solicitada],
                        mocambola=mocambola,
                        config=config,
                        local_only=True,
                    )


bbx = Baobaxia()
