#!/usr/bin/env python3
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional

import shortuuid
from loguru import logger
from pydantic import BaseModel, EmailStr, Extra
from pydantic_settings import BaseSettings
from slugify import slugify

from baobaxia.constants import SABER_FILE_EXT, SLUG_SEPARATOR
from baobaxia.permissions.roles import Role

NoneType = type(None)


class SaberesConfig(BaseSettings):
    """
    Modelo de configuração do Saber.
    Essa classe é usada para armazenar algumas configurações de base do Baobáxia.

    Configuração disponível no arquivo: /home/mocambola/.baobaxia.conf
    """

    default_data_path: Path
    default_balaio: str = "default_balaio"
    default_mucua: str = "default_mucua"
    default_mucua_name: str = "default_mucua_name"


class Saber(BaseModel):
    """Modelo da classe Saber

    Os Saberes são informações memorizadas nas mucuas em formato
    textual e anexo arquivos binarios como imagens, documentos,
    audios, videos.

    :param smid: Id do Saber (SmallID)
    :type smid: str
    :param name: Nome do saber
    :type name: str
    :param slug: Identificativo do Saber, gerado pelo nome e id (Name + Smid)
    :type slug: str
    :param path: Caminho do Saber (Path)
    :type path: str
    :param balaio: Balaio de base do Saber (SLUG)
    :type balaio: str, Optional
    :param mucua: Mucua de base do Saber (SLUG)
    :type mucua: str, Optional
    :param application: Aplicação padrão para este tipo de Saber
    :type application: str
    :param created: Data de gravação
    :type created: datetime
    :param creator: Mocambola gravador
    :type creator: str
    :param last_update: Data do ultimo manejo
    :type last_update: datetime
    :param data: Modelo de dados especifico do Saber (BaseModel).
    :type data: BaseModel, optional

    """

    name: str
    path: Path = Path(".")
    smid: Optional[str] = None
    slug: Optional[str] = None

    balaio_smid: Optional[str] = None
    mucua_smid: Optional[str] = None

    content: List[Dict] = []

    application: str = "root"
    description: Optional[str] = None

    created: Optional[datetime] = None
    creator: Optional[str] = None
    last_update: Optional[datetime] = None

    is_public: bool = False
    _is_local: bool = False  # DADO LOCAL
    _is_requested: bool = False  # DADO LOCAL


class Mocambola(Saber):
    """
    Modelo da classe Mocambola.

    Mocambolas somos nois \\o//

    :param email: Email do mocambola
    :type email: emailStr, optional
    :param mocambo: Mocambo de base (SMID)
    :type mocambo: str, optional
    :param username: Username do mocambola
    :type username: str
    :param name: Nome do mocambola
    :type name: str, optional
    :param is_native: Nativo por padrã não (False)
    :type is_native: str, false
    :param family: Familia do mocambola
    :type family: str, optional

    :param password_hash: Hash da senha
    :type password_hash: str
    :param validation_code: Codigo de validação
    :type validation_code: str

    """

    email: Optional[EmailStr] = None
    mocambo: Optional[str] = None
    username: str
    name: Optional[str] = None
    role: Role = "USER"
    _live_roles: Optional[Dict] = (
        None  # Preenchido a runtime, { balaio_smid : { role: "USER" }}
    )
    permissions: List[str] = []
    password_hash: Optional[str] = None
    recovery_question: Optional[str] = None
    recovery_answer_hash: Optional[str] = None


class MocambolaIn(Saber):
    email: Optional[EmailStr] = None
    mocambo: Optional[str] = None
    path: Path = Path("mocambolas/")
    username: str
    name: Optional[str] = None
    role: Role = "USER"
    permissions: List[str] = []
    password: Optional[str] = None
    recovery_question: Optional[str] = None
    recovery_answer: Optional[str] = None


class Balaio(Saber):
    """
    Modelo da classe Balaio.

    Balaio é o lugar onde guarda as coisas, e também uma pasta gerenciada
    com git-annex.

    """

    default_mucua: Optional[str] = None

    default_ttl: int = 3
    smid_len: int = None
    slug_smid_len: int = None
    slug_name_len: int = None


class Mucua(Saber):
    """
    Modelo da classe Mucua.

    Mucua è um nó da rede Baobáxia, e também é o fruto do Baobá.
    """

    _uri_local: Optional[str] = None
    uri_internet: Optional[str] = None
    uri_path: Optional[str] = None


class Pedido(Saber):
    """
    Modelo da classe Pedido.

    Pedido de copia dos conteudo de um saber.
    """

    path: Path = Path("pedidos")
    ttl: int
    midia_balaio_smid: str
    midia_mucua_smid: str
    midia_smid: str


class Rota(Saber):
    """Modelo da classe de uma Rota .

    Rota è um conjunto de mucua conectado pelos mocambolas que
    percorrem os territorios.

    """

    path: Path = Path("rotas")
    mucuas_smid: List[str] = []
    is_active: bool = False
    praticas: List[str] = []


class SaberIndexed(BaseModel):
    path: Path
    saber: Saber


class SaberesDataStore:
    """
    Classe para armazenar os Saberes em memória e criar objetos de acesso (SaberesDataset).

    :param config: Objeto de configuração SaberConfig
    :type config: SaberConfig

    """

    def __init__(self, config: SaberesConfig):
        """Metodo construtor"""
        super().__init__()
        self.config = config
        self.clear_cache()

    def clear_cache(self):
        """Limpa a cache"""
        self._balaios = {}
        self._saberes = {}

    def cache(self, saber: Saber):
        """Coloca o Saber na cache

        :param saber: Objeto Saber a ser memorizado na cache
        :type saber: Saber

        """
        path = Path(".")
        if saber.balaio_smid is None:
            path = path / saber.path
            self._balaios[saber.smid] = SaberIndexed(path=path, saber=saber)
        else:
            if saber.balaio_smid not in self._saberes:
                self._saberes[saber.balaio_smid] = {}
            if saber.mucua_smid is not None:
                path = (
                    path
                    / self._saberes[saber.balaio_smid][saber.mucua_smid].path
                )
            path = path / saber.path
            self._saberes[saber.balaio_smid][saber.smid] = SaberIndexed(
                path=path, saber=saber
            )

    def uncache(self, smid: str, balaio_smid: Optional[str] = None):
        """Remove o objeto do cache

        :param saber: Objeto Saber a ser removido da cache
        :type saber: Saber

        """
        saber = None
        if balaio_smid is None:
            saber = self._balaios[smid].saber
            del self._balaios[smid]
        else:
            saber = self._saberes[saber.balaio_smid][smid].saber
            del self._saberes[saber.balaio_smid][smid]
        return saber

    def get_cache(self, smid: str, balaio_smid: Optional[str] = None):
        if balaio_smid is None:
            return self._balaios[smid]
        else:
            return self._saberes[balaio_smid][smid]

    def get_cached_saber(self, smid: str, balaio_smid: Optional[str] = None):
        return self.get_cache(smid, balaio_smid).saber

    def get_cached_path(self, smid: str, balaio_smid: Optional[str] = None):
        path = Path(".")
        if balaio_smid is not None:
            path = path / self.get_cached_path(smid=balaio_smid)
        return path / self.get_cache(smid, balaio_smid).path

    def is_cached(self, smid: str, balaio_smid: Optional[str] = None):
        if balaio_smid is None:
            return smid in self._balaios
        else:
            return (
                balaio_smid in self._saberes
                and smid in self._saberes[balaio_smid]
            )

    def create_dataset(
        self,
        model: type,
        balaio_smid: str,
        mucua_smid: str,
        mocambola: Optional[str] = None,
    ):
        return SaberesDataset(self, model, balaio_smid, mucua_smid, mocambola)

    def create_balaio_dataset(self, mocambola: Optional[str] = None):
        return SaberesDataset(self, Balaio, None, None, mocambola)

    def create_mucua_dataset(
        self, balaio_smid: str, mocambola: Optional[str] = None
    ):
        return SaberesDataset(self, Mucua, balaio_smid, None, mocambola)


class SaberesDataset:
    """Classe usada para criar e manejar os saberes."""

    def __init__(
        self,
        datastore: SaberesDataStore,
        model: type,
        balaio_smid: Optional[str] = None,
        mucua_smid: Optional[str] = None,
        mocambola: Optional[str] = None,
    ):
        super().__init__()
        self.datastore = datastore
        self.model = model
        self.balaio_smid = balaio_smid
        self.mucua_smid = mucua_smid
        self.mocambola = mocambola
        self.base_path = None
        self.default_ttl = None
        self.smid_len = None
        self.slug_smid_len = None
        self.slug_name_len = None

        if balaio_smid:
            balaio = self.datastore.get_cached_saber(smid=balaio_smid)
            self._set_keys(balaio)
            # self.default_ttl = balaio.default_ttl
            # self.smid_len = balaio.smid_len
            # self.slug_smid_len = balaio.slug_smid_len
            # self.slug_name_len = balaio.slug_name_len

    def get_base_path(self):
        """Retorna o caminho básico usado pelo Dataset."""
        self.base_path = self.datastore.config.default_data_path
        # self.base_path = self.get(self.balaio_smid).path.parent #FIX pode nao estar setado
        if self.balaio_smid is not None:
            if self.mucua_smid is None:
                self.base_path = (
                    self.base_path
                    / self.datastore.get_cached_path(smid=self.balaio_smid)
                )
            else:
                self.base_path = (
                    self.base_path
                    / self.datastore.get_cached_path(
                        smid=self.mucua_smid, balaio_smid=self.balaio_smid
                    )
                )
        return self.base_path

    def get_dir_path(self, smid: str):
        """Retorna o caminho de um saber (existente ou não)
        dentro do contexto do Dataset.
        """
        return (
            self.datastore.config.default_data_path
            / self.datastore.get_cached_path(smid, self.balaio_smid)
        )

    def get_file_path(self, smid: str):
        """Retorna o caminho do arquivo de metadados
        de um saber (existente ou não)
        dentro do contexto do Dataset.
        """
        bbx_file_ext = SABER_FILE_EXT
        return self.get_dir_path(smid) / bbx_file_ext

    def _set_keys(self, saber):
        self.default_ttl = saber.default_ttl
        self.smid_len = saber.smid_len
        self.slug_smid_len = saber.slug_smid_len
        self.slug_name_len = saber.slug_name_len

    def read_file(self, path: Path):
        """Retorna um saber armazenado em um arquivo."""
        logger.debug(f"SABERES: read_file, Saber Path: {path}")
        saber = self.model.parse_file(path)
        if isinstance(self.model, Balaio):
            self._set_keys(saber)
        saber_local_path = path.with_suffix(".local")
        logger.debug(f"SABERES - read_file, saber.dict: {saber.dict()}")
        if saber_local_path.is_file():
            saber_local = self.model.parse_file(saber_local_path)
            for key in saber.dict():
                if key.startswith("_"):
                    saber.key = saber_local.key
        return saber

    def create_smid(self):
        """Cria um identificador aleatório pequeno (small id)."""
        logger.debug(f"SMID LEN: {self.smid_len}")
        return shortuuid.ShortUUID().random(length=self.smid_len)

    def create_slug(self, smid: str, name: str):
        """Cria uma chave para o saber, usando o identificador
        e o nome atribuído.
        """
        result = slugify(name)
        if len(result) > self.slug_name_len:
            result = result[: self.slug_name_len]
        result += SLUG_SEPARATOR
        if len(smid) > self.slug_smid_len:
            result += smid[: self.slug_smid_len]
        else:
            result += smid
        return result

    def find_and_collect(self, pattern: str):
        """Busca e coleta saberes a partir de um padrão para
        caminhos de arquivos.
        """
        result = []
        baobaxia_files = self.get_base_path().glob(pattern + SABER_FILE_EXT)
        for bf in baobaxia_files:
            result_item = self.read_file(path=bf)
            result.append(result_item.copy())
            self.datastore.cache(result_item)
        return result

    # def conf_and_collect(self, pattern: str,
    #                      abs_path: Path):
    #     """Configura e coleta saberes a partir de um padrão para
    #     caminhos de arquivos e dos arquivos de metadados do baobáxia.

    #     Precisa incluir o link para o arquivo .baobaxia (o outro customizado).
    #     """
    #     result = []

    #     self.datastore.config.saber_file_ext = abs_path.suffix
    #     baobaxia_files = abs_path.parent.glob(
    #         pattern + self.datastore.config.saber_file_ext
    #     )
    #     for bf in baobaxia_files:
    #         result_item = self.read_file(path=bf) # FIX MUDA AQUI O PATH COM O ABS
    #         result.append(result_item.copy())
    #         self.datastore.cache(result_item)
    #     return result

    def collect(self, path: Path):
        """Coleta um saber."""
        result = self.read_file(path)
        self.datastore.cache(result)
        return result.copy()

    def get(self, smid: str):
        """Retorna um saber armazenado em cache."""
        return self.datastore.get_cached_saber(smid, self.balaio_smid)

    def settle(self, saber: Saber, is_syncing=False):
        """Assenta um saber em seu arquivo correspondente."""
        if self.mocambola is None:
            raise RuntimeError("Dataset is readonly (no mocambola defined)")
        if saber.balaio_smid is None:
            saber.balaio_smid = self.balaio_smid
            if isinstance(saber, Balaio):
                self._set_keys(saber)
        if saber.mucua_smid is None:
            saber.mucua_smid = self.mucua_smid
        if saber.smid is not None and self.datastore.is_cached(
            saber.smid, saber.balaio_smid
        ):
            saber_old = self.datastore.get_cached_saber(
                saber.smid, saber.balaio_smid
            )
            saber.created = saber_old.created
            saber.creator = saber_old.creator
            if saber.name != saber_old.name:
                saber.slug = self.create_slug(saber.smid, saber.name)
                saber.path = Path(saber.path).parent / saber.slug
                p = Path(self.get_file_path(saber.smid)).parent
                saberpath = Path(p)
                saberdirpath = Path(saberpath.parent)
                if not saberdirpath.exists():
                    saberdirpath.mkdir(parents=True)
                saberpath.rename(self.get_base_path() / Path(saber.path))
        else:
            if saber.slug is None:
                if saber.smid is None:
                    saber.smid = self.create_smid()
                saber.slug = self.create_slug(saber.smid, saber.name)
                saber.path = saber.path / saber.slug
            saber.created = datetime.now()
            if isinstance(self.mocambola, Mocambola):
                saber.creator = self.mocambola.username
            else:
                saber.creator = self.mocambola
        if not is_syncing:
            saber.last_update = datetime.now()
        self.datastore.cache(saber)
        dirpath = self.get_dir_path(saber.smid)
        if not dirpath.exists():
            dirpath.mkdir(parents=True)
        filepath = self.get_file_path(saber.smid)
        if not filepath.exists():
            filepath.touch()
        filepath.open("w").write(saber.model_dump_json(indent=4))
        if filepath.with_suffix(".local").exists():  # FIX
            saber_local = self.model.parse_file(filepath.with_suffix(".local"))
            for key in saber.dict():
                if key.startswith("_"):
                    saber_local.key = saber.key
            filepath.with_suffix(".local").open("w").write(
                saber.model_dump_json(indent=4)
            )
            return saber_local.copy()
        return saber.copy()

    def drop(self, smid: str):
        """Abandona um saber."""
        if self.mocambola is None:
            raise RuntimeError("Dataset is readonly (no mocambola defined)")
        if self.datastore.is_cached(smid, self.balaio_smid):
            saber = self.datastore.get_cached_saber(smid, self.balaio_smid)
            self.datastore.uncache(saber.smid, self.balaio_smid)
