#!/usr/bin/env python3

import glob
import os
import shutil
import subprocess
from datetime import datetime
from pathlib import Path
from typing import Optional

from datalad.api import Dataset, create
from datalad.config import ConfigManager
from loguru import logger

from baobaxia.constants import SABER_FILE_EXT, SLUG_SEPARATOR

from .saberes import Mocambola, Saber, SaberesConfig, SaberesDataset


class SankofaInfo:
    """Context manager for setting commit info on datalad operations that use it."""

    def __init__(self, balaio, name=None, email=None, scope="local"):
        self.config_manager = ConfigManager(balaio)
        self.email = email if email else "exu@mocambos.net"
        self.name = name if name else "exu"
        self.scope = scope

    def __enter__(self):
        self.config_manager.set("user.email", self.email, self.scope)
        self.config_manager.set("user.name", self.name, self.scope)

    def __exit__(self, exception_type, exception_value, traceback):
        self.config_manager.set("user.email", "exu@mocambos.net", self.scope)
        self.config_manager.set("user.name", "exu", self.scope)


class Sankofa:

    @staticmethod
    def _get_balaio_path(balaio_smid, config):
        p = Path(config.default_data_path)
        balaios = [
            x.name
            for x in p.iterdir()
            if x.is_dir()
            if x.name.split(SLUG_SEPARATOR)[-1] in balaio_smid
        ]
        if len(balaios) > 0:
            return balaios[0]
        return None  # FIX raise exception

    @staticmethod
    def _get_mucua_path(balaio_smid, mucua_smid, config):
        p = Path(config.default_data_path) / Sankofa._get_balaio_path(
            balaio_smid=balaio_smid, config=config
        )
        mucuas = [
            x.name
            for x in p.iterdir()
            if x.is_dir()
            if x.name.split(SLUG_SEPARATOR)[-1] in mucua_smid
        ]
        if len(mucuas) > 0:
            return mucuas[0]
        return None  # FIX raise exception

    @staticmethod
    def _prepare_paths(saberes: [Saber], config: SaberesConfig):
        sankofa_data = {}
        balaios = {b.balaio_smid for b in saberes if not b.balaio_smid is None}
        for balaio_smid in balaios:
            saberes_paths = []
            balaio_path = Path(".")
            for saber in saberes:
                if saber.balaio_smid == balaio_smid:
                    balaio_path = Sankofa._get_balaio_path(
                        balaio_smid=balaio_smid, config=config
                    )
                    logger.debug(f"balaio_path: {balaio_path}")
                    # balaio_path: rede-mocambos-acervo_3KmBEssEi
                    if saber.mucua_smid:
                        saber_local_path = Sankofa._get_mucua_path(
                            balaio_smid=balaio_smid,
                            mucua_smid=saber.mucua_smid,
                            config=config,
                        ) / Path(saber.path)
                    else:
                        saber_local_path = Path(saber.path)
                    logger.debug(f"saber_local_path, {saber_local_path}")
                    # saber_local_path, abdias-nascimento_ZPsa6czJU/acervo/imagens/amores_HpQijdyKr
                    saberes_paths.append(saber_local_path)
                    saberes_paths.append(saber_local_path / SABER_FILE_EXT)
                    for content_path in saber.content:
                        saberes_paths.append(
                            saber_local_path / content_path["filename"]
                        )
            sankofa_data[balaio_path] = saberes_paths
        logger.debug(f"sankofa_data: {sankofa_data}")
        # sankofa_data: {'rede-mocambos-acervo_3KmBEssEi':
        #                [
        #                    PosixPath('abdias-nascimento_ZPsa6czJU/acervo/imagens/amores_HpQijdyKr'),
        #                    PosixPath('abdias-nascimento_ZPsa6czJU/acervo/imagens/amores_HpQijdyKr/.baobaxia'),
        #                    PosixPath('abdias-nascimento_ZPsa6czJU/acervo/imagens/amores_HpQijdyKr/IMG_20200426_162538875_HDR.jpg')
        #                ]
        #                }
        return sankofa_data

    @classmethod
    def create_balaio(
        cls, *, balaio: str, description: str, config: SaberesConfig
    ):
        balaio_path = Path(balaio)
        balaio_fullpath = config.default_data_path / balaio_path
        dataset = create(
            path=balaio_fullpath,
            force=True,
            annex=True,
            cfg_proc="text2git",
            description=description,
        )
        balaio_ignore_path = balaio_fullpath.joinpath(".gitignore")
        with open(balaio_ignore_path, "w") as balaio_ignore:
            balaio_ignore.write("**.local" + "\n" + "**_thumb.*" + "\n")
        balaio_attributes_path = balaio_fullpath.joinpath(".gitattributes")
        with open(balaio_attributes_path, "a") as balaio_atttributes:
            balaio_atttributes.write("**/*.md annex.largefiles=anything")
        dataset.save()

    @classmethod
    def destroy_balaio(cls, balaio: str, config: SaberesConfig):
        shutil.rmtree(config.default_data_path / balaio)

    # Metodos para atualizar os dados da Mucua, rotas, ssh e territorios

    @classmethod
    def mucua_remote_add(
        cls,
        *,
        balaio_slug: str,
        mucua_slug: str,
        uri: str,
        config: SaberesConfig,
    ):
        repopath = Path(config.default_data_path)
        balaiopath = repopath / balaio_slug

        cmd = "git remote add " + mucua_slug + " " + uri
        pipe = subprocess.Popen(cmd, shell=True, cwd=balaiopath)
        pipe.wait()

    @classmethod
    def mucua_remote_get(cls, *, balaio_slug: str, config: SaberesConfig):
        repopath = Path(config.default_data_path)
        balaiopath = repopath / balaio_slug
        cmd = "git remote -v"
        pipe = subprocess.Popen(
            cmd, shell=True, cwd=balaiopath, stdout=subprocess.PIPE
        )
        output, error = pipe.communicate()

        mucuas = []
        # Match repositories.
        if output:
            for line in output.splitlines():
                mucuas.append(line.split(None, 1)[1].split()[0].decode("utf-8"))
        return list(set(mucuas))

    @classmethod
    def mucua_remote_del(
        cls, *, balaio_slug: str, mucua_slug: str, config: SaberesConfig
    ):
        repopath = Path(config.default_data_path)
        balaiopath = repopath / balaio_slug
        cmd = "git remote rm " + mucua_slug
        pipe = subprocess.Popen(cmd, shell=True, cwd=balaiopath)
        pipe.wait()

    @classmethod
    def mucua_group_add(
        cls,
        *,
        balaio_slug: str,
        mucua_slug: str,
        group: str,
        config: SaberesConfig,
    ):
        repopath = Path(config.default_data_path)
        balaiopath = repopath / balaio_slug
        cmd = "git annex group " + mucua_slug + " " + group
        pipe = subprocess.Popen(
            cmd, shell=True, cwd=balaiopath, stdout=subprocess.PIPE
        )
        output, error = pipe.communicate()

    @classmethod
    def mucua_group_del(
        cls,
        *,
        balaio_slug: str,
        mucua_slug: str,
        group: str,
        config: SaberesConfig,
    ):
        repopath = Path(config.default_data_path)
        balaiopath = repopath / balaio_slug
        cmd = "git annex ungroup " + mucua_slug + " " + group
        pipe = subprocess.Popen(
            cmd, shell=True, cwd=balaiopath, stdout=subprocess.PIPE
        )
        output, error = pipe.communicate()

    @classmethod
    def mucua_group_list(
        cls, *, balaio_slug: str, mucua_slug: str = None, config: SaberesConfig
    ):
        if mucua_slug == None:
            mucua_slug = config.mucua_slug

        repopath = Path(config.default_data_path)
        balaiopath = repopath / balaio_slug
        cmd = "git annex group " + mucua_slug
        pipe = subprocess.Popen(
            cmd, shell=True, cwd=balaiopath, stdout=subprocess.PIPE
        )
        output, error = pipe.communicate()
        if output != "":
            return output.decode("utf-8").split()
        else:
            return []

    @classmethod
    def set_remote(cls, *, balaio_path, name, uri, config: SaberesConfig):

        repopath = Path(config.default_data_path)
        balaiopath = Path(repopath) / Path(balaio_path)
        dataset = Dataset(path=balaiopath)

        with SankofaInfo(balaio=dataset):
            dataset.configuration("set", [("remote." + name + ".url", uri)])

    @classmethod
    def get(cls, *, saberes: [Saber], config: SaberesConfig):

        sankofa_data = Sankofa._prepare_paths(saberes, config)

        for balaio, saberes_paths in sankofa_data.items():
            repopath = Path(config.default_data_path)
            balaiopath = Path(repopath) / Path(balaio)
            dataset = Dataset(path=balaiopath)

            with SankofaInfo(balaio=dataset):
                dataset.get(path=saberes_paths)

    @classmethod
    def add(
        cls, *, saberes: [Saber], mocambola: Mocambola, config: SaberesConfig
    ):
        sankofa_data = Sankofa._prepare_paths(saberes=saberes, config=config)

        for balaio, saberes_paths in sankofa_data.items():
            repopath = Path(config.default_data_path)
            balaiopath = Path(repopath) / Path(balaio)
            dataset = Dataset(path=balaiopath)

            with SankofaInfo(
                balaio=dataset, name=mocambola.username, email=mocambola.email
            ):
                return dataset.save(path=saberes_paths)

    @classmethod
    def remove(
        cls,
        *,
        saberes: [Saber],
        mocambola: Mocambola,
        config: SaberesConfig,
        local_only=False,
    ):
        """
        Remove os files associados aos saberes de forma permanente em todas
        as mucuas (se local_only=False (padrão)) ou apenas localmente (se
        local_only=True).

        """
        sankofa_data = Sankofa._prepare_paths(saberes, config)

        for balaio, saberes_paths in sankofa_data.items():
            repopath = Path(config.default_data_path)
            balaiopath = Path(repopath) / Path(balaio)
            dataset = Dataset(path=balaiopath)

            with SankofaInfo(
                balaio=dataset, name=mocambola.username, email=mocambola.email
            ):
                try:
                    if local_only:
                        dataset.drop(
                            path=saberes_paths, recursive=True, check=False
                        )
                    else:
                        dataset.remove(
                            path=saberes_paths, recursive=True, check=False
                        )
                except:
                    # Caso o arquivo não esteja no git (TODO especializar except)
                    for saber in saberes_paths:
                        saberpath = balaiopath / saber
                        if saberpath.exists() and saberpath.is_dir():
                            shutil.rmtree(saberpath)
                        else:
                            print(
                                "Sankofa.remove (not found): " + str(saberpath)
                            )

    @classmethod
    def sync(
        cls,
        *,
        balaio_slug: Optional[str] = None,
        mucua_slug: Optional[str] = None,
        config: SaberesConfig,
        **kwargs,
    ):
        """
        Sincroniza com outra mucua.
        """

        repopath = config.default_data_path
        balaiopath = repopath / balaio_slug
        dataset = Dataset(balaiopath)

        with SankofaInfo(balaio=dataset):
            dataset.update(**kwargs)
