#!/usr/bin/env python3

import json
from hashlib import md5
from mimetypes import guess_type
from pathlib import Path
from shutil import copy

from loguru import logger
from passlib.context import CryptContext
from pydantic import BaseModel
from unidecode import unidecode


class GeoLocation(BaseModel):
    latitude: float = 0.0
    longitude: float = 0.0
    description: str = ""


def create_link(src: str, target: str):
    if not target.exists():
        copy(src, target, follow_symlinks=False)


def copy_file(src: str, target: str):
    if not target.exists():
        copy(src, target, follow_symlinks=True)


def calculate_md5(file_path: str) -> str:
    md5_hash = md5()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            md5_hash.update(chunk)
    return md5_hash.hexdigest()


def str_to_hash(base: str):
    pwd_context = CryptContext(schemes=["bcrypt"], deprecated="auto")
    hash_pass = pwd_context.hash(base)
    return hash_pass


def norm(data):
    if isinstance(data, list):
        normlist = []
        for value in data:
            normlist.append(unidecode(value.casefold()))
        return normlist
    elif isinstance(data, str):
        return unidecode(data.casefold())
    else:
        return data


def rm_duplicate_dict_in_list(dlist: list):
    visto = set()
    nova_lista = []
    for d in dlist:
        t = tuple(sorted(d.items()))
        if t not in visto:
            visto.add(t)
            nova_lista.append(d)
    return nova_lista


def processar_arquivo_json(
    caminho, atributo, novo_valor, excluidos=None, incluidos=None
):
    caminho = Path(caminho)
    if excluidos is None:
        excluidos = []
    if incluidos is None:
        incluidos = []
    logger.debug(f"PROCESSAR_ARQUIVO_JSON, CAMINHO: {caminho}")
    if caminho.is_file() and caminho.name == ".baobaxia":
        if all(excluido not in str(caminho) for excluido in excluidos) and all(
            incluido in str(caminho) for incluido in incluidos
        ):
            with open(caminho, "r", encoding="utf-8") as arquivo:
                try:
                    dados = json.load(arquivo)
                    logger.debug(f"PROCESSAR_ARQUIVO_JSON, DADOS: {dados}")
                except json.JSONDecodeError:
                    print(f"Erro ao ler o arquivo Baobaxia: {caminho}")
                    return
                for chave, valor in dados.items():
                    if chave == atributo:
                        dados[chave] = novo_valor
            # Reescrevemos o arquivo com os novos valores
            with open(caminho, "w", encoding="utf-8") as arquivo:
                json.dump(dados, arquivo, indent=4)

    elif caminho.is_dir():
        # Se for um diretório, percorremos os arquivos e pastas dentro dele
        for item in caminho.iterdir():
            if all(excluido not in str(item) for excluido in excluidos) and all(
                incluido in str(item) for incluido in incluidos
            ):
                processar_arquivo_json(
                    item, atributo, novo_valor, excluidos, incluidos
                )


def processar_arquivo_json_remove_attribute(
    caminho, atributo, excluidos=None, incluidos=None
):
    caminho = Path(caminho)
    if excluidos is None:
        excluidos = []
    if incluidos is None:
        incluidos = [""]
    logger.debug(f"PROCESSAR_ARQUIVO_JSON, CAMINHO: {caminho}")
    if caminho.is_file() and caminho.name == ".baobaxia":
        if all(excluido not in str(caminho) for excluido in excluidos) and all(
            incluido in str(caminho) for incluido in incluidos
        ):
            with open(caminho, "r", encoding="utf-8") as arquivo:
                try:
                    dados = json.load(arquivo)
                    logger.debug(f"PROCESSAR_ARQUIVO_JSON, DADOS: {dados}")
                except json.JSONDecodeError:
                    print(f"Erro ao ler o arquivo Baobaxia: {caminho}")
                    return

                for key, value in dados.copy().items():
                    if key == atributo:
                        dados.pop(key)
                    logger.debug(f"DADOS agora: {dados}")
            # Reescrevemos o arquivo com os novos valores
            with open(caminho, "w", encoding="utf-8") as arquivo:
                json.dump(dados, arquivo, indent=4)

    elif caminho.is_dir():
        # Se for um diretório, percorremos os arquivos e pastas dentro dele
        for item in caminho.iterdir():
            if all(excluido not in str(item) for excluido in excluidos) and all(
                incluido in str(item) for incluido in incluidos
            ):
                processar_arquivo_json_remove_attribute(
                    item, atributo, excluidos, incluidos
                )


def processar_arquivo_json_indent(caminho, excluidos=None, incluidos=None):
    caminho = Path(caminho)
    if excluidos is None:
        excluidos = []
    if incluidos is None:
        incluidos = [""]
    logger.debug(f"PROCESSAR_ARQUIVO_JSON, CAMINHO: {caminho}")
    if caminho.is_file() and caminho.name == ".baobaxia":
        if all(excluido not in str(caminho) for excluido in excluidos) and all(
            incluido in str(caminho) for incluido in incluidos
        ):
            with open(caminho, "r", encoding="utf-8") as arquivo:
                try:
                    dados = json.load(arquivo)
                    logger.debug(f"PROCESSAR_ARQUIVO_JSON, DADOS: {dados}")
                except json.JSONDecodeError:
                    print(f"Erro ao ler o arquivo Baobaxia: {caminho}")
                    return

            # Reescrevemos o arquivo com os novos valores
            with open(caminho, "w", encoding="utf-8") as arquivo:
                json.dump(dados, arquivo, indent=4)

    elif caminho.is_dir():
        # Se for um diretório, percorremos os arquivos e pastas dentro dele
        for item in caminho.iterdir():
            if all(excluido not in str(item) for excluido in excluidos) and all(
                incluido in str(item) for incluido in incluidos
            ):
                processar_arquivo_json_indent(item, excluidos, incluidos)


def discover_mimetype(filename):
    try:
        mimetype, encoding = guess_type(filename)
    except Exception as e:
        logger.error(f"Erro ao tentar obter mime type: {e}")
    return mimetype
