import os
from datetime import datetime
from pathlib import Path
from typing import Optional

import pytest
from datalad.api import create
from pydantic import BaseModel

from baobaxia.saberes import (
    Balaio,
    Mocambola,
    Mucua,
    Saber,
    SaberesConfig,
    SaberesDataset,
    SaberesDataStore,
)
from baobaxia.util import str_to_hash


class DataTestModel(Saber):
    message: str
    any_int: int = -1


class BaseBaobaxiaTest:

    data_path = Path("/tmp/baobaxia_tests")
    mocambolas_path = Path("mocambolas")
    tests_path = Path("tests")

    default_date = datetime(2021, 3, 29, 0, 0, 0, 0)

    balaio_name = "meu-balaio"
    balaio_smid = "5atXevYw2"
    balaio_slug = balaio_name + "_" + balaio_smid

    mucua_name = "abdias"
    mucua_smid = "4jtRevYwk"
    mucua_slug = mucua_name + "_" + mucua_smid

    mocambola_smid = "6atRevFwk"
    mocambola_username = "exu"
    mocambola_slug = mocambola_username + "_" + mocambola_smid
    mocambola_path = mocambolas_path / mocambola_slug
    mocambola_email = "exu@mocambos.net"
    mocambola_password = "livre123"

    tests_app = "bbx_test"

    config = SaberesConfig(
        data_path=data_path,
        default_balaio=balaio_smid,
        default_mucua=mucua_smid,
    )

    balaio = Balaio(
        name=balaio_name,
        smid=balaio_smid,
        slug=balaio_slug,
        path=Path(balaio_slug),
        creator=mocambola_username,
        created=default_date,
        last_update=default_date,
        default_mucua=mucua_smid,
    )

    mucua = Mucua(
        name=mucua_name,
        smid=mucua_smid,
        slug=mucua_slug,
        path=Path(mucua_slug),
        balaio_smid=balaio_smid,
        creator=mocambola_username,
        created=default_date,
        last_update=default_date,
    )

    mocambola = Mocambola(
        smid=mocambola_smid,
        name=mocambola_username,
        slug=mocambola_slug,
        path=mocambola_path,
        balaio_smid=balaio_smid,
        mucua_smid=mucua_smid,
        creator=mocambola_username,
        created=default_date,
        last_update=default_date,
        email=mocambola_email,
        username=mocambola_username,
        roles=[],
        password_hash=str_to_hash(mocambola_password),
    )

    saber_dir_name = "tests"
    saber_dir_message = "Pasta de testes"
    saber_dir_any_int = 1

    saber_file1_name = "olamundo"
    saber_file1_message = "Olá, mundo!!!"
    saber_file1_any_int = 11
    saber_file1_content = Path("olamundo.txt")

    saber_file2_name = "ciaomundo"
    saber_file2_message = "Ciao, mundo!!!"
    saber_file2_any_int = 12
    saber_file2_content = Path("ciaomundo.bin")

    def init_balaio_local(self):
        # BALAIO de base
        if not self.data_path.exists():
            self.data_path.mkdir()

        self.datastore = SaberesDataStore(self.config)

        self.balaio_path = self.data_path / self.balaio_slug
        self.mucua_path = self.balaio_path / self.mucua_slug

        if not self.balaio_path.exists():
            dataset = self.datastore.create_balaio_dataset(
                mocambola=self.mocambola_username
            )
            dataset.settle(self.balaio)

        if not self.mucua_path.exists():
            dataset = self.datastore.create_mucua_dataset(
                balaio_smid=self.balaio_smid, mocambola=self.mocambola_username
            )
            dataset.settle(self.mucua)

        mocambolas_fullpath = (
            self.data_path
            / self.balaio.slug
            / self.mucua.slug
            / self.mocambolas_path
        )
        if not mocambolas_fullpath.exists():
            mocambolas_fullpath.mkdir()

        dataset = self.datastore.create_dataset(
            model=Mocambola,
            balaio_smid=self.balaio_smid,
            mucua_smid=self.mucua_smid,
            mocambola=self.mocambola_username,
        )
        dataset.settle(self.mocambola)

        tests_fullpath = (
            self.data_path
            / self.balaio_slug
            / self.mucua_slug
            / self.tests_path
        )
        if not tests_fullpath.exists():
            tests_fullpath.mkdir()

        # Inicializa um repositorio git temporariamente e adiciona no
        # .git/config o UUID da Mucua Local
        create(path=self.balaio_path, annex=False, force=True)
        gitconfig = self.balaio_path / ".git" / "config"
        os.remove(self.balaio_path / ".noannex")
        with open(gitconfig, "a") as f:
            f.write("[annex]\n")
            f.write("\tuuid = " + str(self.mucua_smid) + "\n")
        # Agora reinicializa de verdade com os parametros certos.
        dataset = create(
            path=self.balaio_path,
            force=True,
            annex=True,
            cfg_proc="text2git",
            description=self.mucua_slug,
        )
        dataset.save()
